from ariths_gen.wire_components import (
    ConstantWireValue0,
    ConstantWireValue1,
    Bus
)
from ariths_gen.core.arithmetic_circuits import (
    MultiplierCircuit
)
from ariths_gen.one_bit_circuits.one_bit_components import (
    HalfAdder,
    FullAdder
)
from ariths_gen.one_bit_circuits.logic_gates import (
    AndGate,
    NandGate,
    NorGate
)
from ariths_gen.multi_bit_circuits.adders import (
    UnsignedCarryLookaheadAdder
)


class UnsignedCarrySaveMultiplier(MultiplierCircuit):
    """Class representing unsigned carry save array multiplier (also known as Braun multiplier).

    Unsigned carry save array multiplier represents N-bit multiplier composed of
    many AND gates and carry save adders to calculate partial products and
    gradually sum them. As opposed to traditional architecture of an array multiplier,
    the individual half/full adders are interconnected in a carry save manner to provide
    smaller propagation delay.

    It is composed of many logic gates and its downside is a rather big area when compared
    to a regular array multiplier.
    The reason for that is because the csa multiplier requires an additional vector merging
    propagate adder to calculate the final product bits.
    ```
                                        A3B0     A2B0     A1B0     A0B0
                                        │ │      │ │      │ │      │ │
                                       ┌▼─▼┐    ┌▼─▼┐    ┌▼─▼┐    ┌▼─▼┐
                                       │AND│    │AND│    │AND│    │AND│
                                       └┬──┘    └┬──┘    └┬──┘    └─┬─┘
                                  A3B1  │  A2B1  │  A1B1  │  A0B1   │
                                 ┌▼─▼┐  │ ┌▼─▼┐  │ ┌▼─▼┐  │ ┌▼─▼┐   │
                                 │AND│  │ │AND│  │ │AND│  │ │AND│   │
                                 └┬──┘  │ └┬──┘  │ └┬──┘  │ └┬──┘   │
                                  │     │  │     │  │     │  │      │
                                  │    ┌▼──▼┐   ┌▼──▼┐   ┌▼──▼┐     │
                               ┌──┘    │    │   │    │   │    │     │
                               │     ┌─┤ HA │ ┌─┤ HA │ ┌─┤ HA │     │
                               │     │ │    │ │ │    │ │ │    │     │
                               │     │ └┬───┘ │ └┬───┘ │ └─┬──┘     │
                        A3B2   │ A2B2│  │ A1B2│  │ A0B2│   │        │
                       ┌▼─▼┐   │┌▼─▼┐│  │┌▼─▼┐│  │┌▼─▼┐│   │        │
                       │AND│   ││AND││  ││AND││  ││AND││   │        │
                       └─┬─┘   │└─┬─┘│  │└─┬─┘│  │└─┬─┘│   │        │
                         │     │  │  │  │  │  │  │  │  │   │        │
                         │    ┌▼──▼┐ │ ┌▼──▼┐ │ ┌▼──▼┐ │   │        │
                      ┌──┘    │    │◄┘ │    │◄┘ │    │◄┘   │        │
                      │     ┌─┤ FA │ ┌─┤ FA │ ┌─┤ FA │     │        │
                      │     │ │    │ │ │    │ │ │    │     │        │
                      │     │ └┬───┘ │ └┬───┘ │ └─┬──┘     │        │
               A3B3   │ A2B3│  │ A1B3│  │ A0B3│   │        │        │
              ┌▼─▼┐   │┌▼─▼┐│  │┌▼─▼┐│  │┌▼─▼┐│   │        │        │
              │AND│   ││AND││  ││AND││  ││AND││   │        │        │
              └─┬─┘   │└─┬─┘│  │└─┬─┘│  │└─┬─┘│   │        │        │
                │     │  │  │  │  │  │  │  │  │   │        │        │
                │    ┌▼──▼┐ │ ┌▼──▼┐ │ ┌▼──▼┐ │   │        │        │
              ┌─┘    │    │◄┘ │    │◄┘ │    │◄┘   │        │        │
              │    ┌─┤ FA │ ┌─┤ FA │ ┌─┤ FA │     │        │        │
              │    │ │    │ │ │    │ │ │    │     │        │        │
      0   0   │    │ └─┬──┘ │ └─┬──┘ │ └─┬──┘     │        │        │
      │   │   │S2  │C2 │S1  │C1 │S0  │C0 │        │        │        │
    ┌─▼───▼───▼────▼───▼────▼───▼────▼─┐ │        │        │        │
    │         Carry-propagate          │ │        │        │        │
    │              adder               │ │        │        │        │
    └┬────────┬────────┬────────┬──────┘ │        │        │        │
     │        │        │        │        │        │        │        │
     ▼        ▼        ▼        ▼        ▼        ▼        ▼        ▼
     P7       P6       P5       P4       P3       P2       P1       P0
    ```

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        prefix (str, optional): Prefix name of unsigned carry save array multiplier. Defaults to "".
        name (str, optional): Name of unsigned carry save array multiplier. Defaults to "u_csamul".
        unsigned_adder_class_name (str, optional): Unsigned multi bit adder used for final vector merging of sums and carries. Defaults to UnsignedCarryLookaheadAdder.
    """
    def __init__(self, a: Bus, b: Bus, prefix: str = "", name: str = "u_csamul", unsigned_adder_class_name: str = UnsignedCarryLookaheadAdder, **kwargs):
        self.N = max(a.N, b.N)
        super().__init__(inputs=[a, b], prefix=prefix, name=name, out_N=self.N*2, **kwargs)

        # Bus sign extension in case buses have different lengths
        self.a.bus_extend(N=self.N, prefix=a.prefix)
        self.b.bus_extend(N=self.N, prefix=b.prefix)

        # Gradual generation of partial products
        for b_multiplier_index in range(self.N):
            for a_multiplicand_index in range(self.N):
                # AND gates generation for calculation of partial products
                obj_and = AndGate(self.a.get_wire(a_multiplicand_index), self.b.get_wire(b_multiplier_index), prefix=self.prefix+"_and"+str(a_multiplicand_index)+"_"+str(b_multiplier_index))
                self.add_component(obj_and)

                # First row contains only HAs
                if b_multiplier_index == 1 and a_multiplicand_index != self.N-1:
                    previous_product = self.components[a_multiplicand_index + b_multiplier_index].out

                    obj_adder = HalfAdder(a=obj_and.out, b=previous_product, prefix=self.prefix+"_ha"+str(a_multiplicand_index)+"_"+str(b_multiplier_index))
                    self.add_component(obj_adder)
                # Rest are composed of FAs
                elif b_multiplier_index > 1 and a_multiplicand_index != self.N-1:
                    previous_sum_wire = self.get_previous_component((self.N-1)*2-1).get_sum_wire() if a_multiplicand_index != self.N-2 else self.get_previous_component((self.N-1)*2).out
                    previous_carry_wire = self.get_previous_component(self.N*2-1).get_carry_wire()

                    obj_adder = FullAdder(a=obj_and.out, b=previous_sum_wire, c=previous_carry_wire, prefix=self.prefix+"_fa"+str(a_multiplicand_index)+"_"+str(b_multiplier_index))
                    self.add_component(obj_adder)

                # PRODUCT GENERATION
                if a_multiplicand_index == 0 and b_multiplier_index == 0:
                    self.out.connect(a_multiplicand_index, obj_and.out)

                    # 1 bit multiplier case
                    if a_multiplicand_index == self.N-1:
                        self.out.connect(a_multiplicand_index+1, ConstantWireValue0())
                        return

                elif a_multiplicand_index == 0:
                    self.out.connect(b_multiplier_index + a_multiplicand_index, obj_adder.get_sum_wire())

        # Final addition of remaining bits using chosen unsigned multi bit adder
        previous_sums = [self.get_previous_component((self.N-2-wire_id)*2).get_sum_wire() if wire_id < self.N-2 else self.get_previous_component().out if wire_id == self.N-2 else ConstantWireValue0() for wire_id in range(self.N)]
        previous_carries = [self.get_previous_component((self.N-2-wire_id)*2+2).get_carry_wire() if wire_id != self.N-1 else ConstantWireValue0() for wire_id in range(self.N)]

        # Obtain proper adder name with its bit width (columns bit pairs minus the first alone bit)
        adder_name = unsigned_adder_class_name(a=a, b=b).prefix + str(self.N)
        adder_a = Bus(prefix=f"a", wires_list=previous_sums)
        adder_b = Bus(prefix=f"b", wires_list=previous_carries)

        final_adder = unsigned_adder_class_name(a=adder_a, b=adder_b, prefix=self.prefix, name=adder_name, inner_component=True, **kwargs)
        self.add_component(final_adder)
        [self.out.connect(o, final_adder.out.get_wire(o-self.N), inserted_wire_desired_index=o-self.N) for o in range(self.N, len(self.out.bus))]


class SignedCarrySaveMultiplier(MultiplierCircuit):
    """Class representing signed carry save array multiplier.

    Signed carry save array multiplier represents N-bit multiplier composed of
    many AND/NAND gates and carry save adders to calculate partial products and
    gradually sum them. As opposed to traditional architecture of an array multiplier,
    the individual half/full adders are interconnected in a carry save manner to provide
    smaller propagation delay.

    It is composed of many logic gates and its downside is a rather big area when compared
    to a regular array multiplier.
    The reason for that is because the csa multiplier requires an additional vector merging
    propagate adder to calculate the final product bits.
    ```
                                        A3B0     A2B0     A1B0     A0B0
                                        │ │      │ │      │ │      │ │
                                       ┌▼─▼─┐   ┌▼─▼┐    ┌▼─▼┐    ┌▼─▼┐
                                       │NAND│   │AND│    │AND│    │AND│
                                       └┬───┘   └┬──┘    └┬──┘    └─┬─┘
                                  A3B1  │  A2B1  │  A1B1  │  A0B1   │
                                 ┌▼─▼─┐ │ ┌▼─▼┐  │ ┌▼─▼┐  │ ┌▼─▼┐   │
                                 │NAND│ │ │AND│  │ │AND│  │ │AND│   │
                               1 └┬───┘ │ └┬──┘  │ └┬──┘  │ └┬──┘   │
                               │  │     │  │     │  │     │  │      │
                              ┌▼──▼┐   ┌▼──▼┐   ┌▼──▼┐   ┌▼──▼┐     │
                              │    │   │    │   │    │   │    │     │
                            ┌─┤ HA │ ┌─┤ HA │ ┌─┤ HA │ ┌─┤ HA │     │
                            │ │    │ │ │    │ │ │    │ │ │    │     │
                            │ └┬───┘ │ └┬───┘ │ └┬───┘ │ └─┬──┘     │
                       A3B2 │  │ A2B2│  │ A1B2│  │ A0B2│   │        │
                      ┌▼─▼─┐│  │┌▼─▼┐│  │┌▼─▼┐│  │┌▼─▼┐│   │        │
                      │NAND││  ││AND││  ││AND││  ││AND││   │        │
                      └──┬─┘│  │└─┬─┘│  │└─┬─┘│  │└─┬─┘│   │        │
                         │  │  │  │  │  │  │  │  │  │  │   │        │
                     ┌───▼┐ │ ┌▼──▼┐ │ ┌▼──▼┐ │ ┌▼──▼┐ │   │        │
                     │    │◄┘ │    │◄┘ │    │◄┘ │    │◄┘   │        │
                   ┌─┤ HA │  ┌┤ FA │  ┌┤ FA │  ┌┤ FA │     │        │
                   │ │    │  ││    │  ││    │  ││    │     │        │
                   │ └┬───┘  │└┬───┘  │└┬───┘  │└─┬──┘     │        │
               A3B3│  │ A3B2 │ │ A3B2 │ │ A3B2 │  │        │        │
              ┌▼─▼┐│  │┌▼─▼─┐│ │┌▼─▼─┐│ │┌▼─▼─┐│  │        │        │
              │AND││  ││NAND││ ││NAND││ ││NAND││  │        │        │
              └─┬─┘│  │└─┬──┘│ │└─┬──┘│ │└─┬──┘│  │        │        │
                │  │  │  │   │ │  │   │ │  │   │  │        │        │
            ┌───▼┐ │ ┌▼──▼┐  │┌▼──▼┐  │┌▼──▼┐  │  │        │        │
            │    │◄┘ │    │◄─┘│    │◄─┘│    │◄─┘  │        │        │
          ┌─┤ HA │ ┌─┤ FA │ ┌─┤ FA │ ┌─┤ FA │     │        │        │
          │ │    │ │ │    │ │ │    │ │ │    │     │        │        │
      1   │ └─┬──┘ │ └─┬──┘ │ └─┬──┘ │ └─┬──┘     │        │        │
      │   │C3 │S2  │C2 │S1  │C1 │S0  │C0 │        │        │        │
    ┌─▼───▼───▼────▼───▼────▼───▼────▼─┐ │        │        │        │
    │         Carry-propagate          │ │        │        │        │
    │              adder               │ │        │        │        │
    └┬────────┬────────┬────────┬──────┘ │        │        │        │
     │        │        │        │        │        │        │        │
     ▼        ▼        ▼        ▼        ▼        ▼        ▼        ▼
     P7       P6       P5       P4       P3       P2       P1       P0
    ```

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        prefix (str, optional): Prefix name of signed carry save array multiplier. Defaults to "".
        name (str, optional): Name of signed carry save array multiplier. Defaults to "s_csamul".
        unsigned_adder_class_name (str, optional): Unsigned multi bit adder used for final vector merging of sums and carries. Defaults to UnsignedCarryLookaheadAdder.
    """
    def __init__(self, a: Bus, b: Bus, prefix: str = "", name: str = "s_csamul", unsigned_adder_class_name: str = UnsignedCarryLookaheadAdder, **kwargs):
        self.N = max(a.N, b.N)
        super().__init__(inputs=[a, b], prefix=prefix, name=name, out_N=self.N*2, signed=True, **kwargs)

        # Bus sign extension in case buses have different lengths
        self.a.bus_extend(N=self.N, prefix=a.prefix, desired_extension_wire=self.a.get_wire(self.a.N-1))
        self.b.bus_extend(N=self.N, prefix=b.prefix, desired_extension_wire=self.b.get_wire(self.b.N-1))

        # Gradual generation of partial products
        for b_multiplier_index in range(self.N):
            for a_multiplicand_index in range(self.N):
                # AND and NAND gates generation for calculation of partial products and sign extension
                if (b_multiplier_index == self.N-1 and a_multiplicand_index != self.N-1) or (b_multiplier_index != self.N-1 and a_multiplicand_index == self.N-1):
                    obj_gate = NandGate(self.a.get_wire(a_multiplicand_index), self.b.get_wire(b_multiplier_index), prefix=self.prefix+"_nand"+str(a_multiplicand_index)+"_"+str(b_multiplier_index))
                    self.add_component(obj_gate)
                else:
                    obj_gate = AndGate(self.a.get_wire(a_multiplicand_index), self.b.get_wire(b_multiplier_index), prefix=self.prefix+"_and"+str(a_multiplicand_index)+"_"+str(b_multiplier_index))
                    self.add_component(obj_gate)

                # First row contains only half adders
                if b_multiplier_index == 1:
                    previous_product = self.components[a_multiplicand_index + b_multiplier_index].out if a_multiplicand_index != self.N-1 else ConstantWireValue1()

                    obj_adder = HalfAdder(a=obj_gate.out, b=previous_product, prefix=self.prefix+"_ha"+str(a_multiplicand_index)+"_"+str(b_multiplier_index))
                    self.add_component(obj_adder)
                # Rest are composed of FAs and a HA
                elif b_multiplier_index > 1:
                    previous_sum_wire = self.get_previous_component((self.N-1)*2).get_sum_wire() if a_multiplicand_index != self.N-1 else None
                    previous_carry_wire = self.get_previous_component(self.N*2).get_carry_wire()

                    # Last function block of each row is a simple HA
                    if a_multiplicand_index == self.N-1:
                        obj_adder = HalfAdder(a=obj_gate.out, b=previous_carry_wire, prefix=self.prefix+"_ha"+str(a_multiplicand_index)+"_"+str(b_multiplier_index))
                    # Rest are all FAs
                    else:
                        obj_adder = FullAdder(a=obj_gate.out, b=previous_sum_wire, c=previous_carry_wire, prefix=self.prefix+"_fa"+str(a_multiplicand_index)+"_"+str(b_multiplier_index))
                    self.add_component(obj_adder)

                # PRODUCT GENERATION
                if a_multiplicand_index == 0 and b_multiplier_index == 0:
                    self.out.connect(a_multiplicand_index, obj_gate.out)

                    # 1 bit multiplier case
                    if a_multiplicand_index == self.N-1:
                        obj_nor = NorGate(a=ConstantWireValue1(), b=self.get_previous_component().out, prefix=self.prefix+"_nor_zero_extend", parent_component=self)
                        self.add_component(obj_nor)

                        self.out.connect(a_multiplicand_index+1, obj_nor.out)
                        return

                elif a_multiplicand_index == 0:
                    self.out.connect(b_multiplier_index + a_multiplicand_index, obj_adder.get_sum_wire())

        # Final addition of remaining bits using chosen unsigned multi bit adder
        previous_sums = [self.get_previous_component((self.N*2)-3-(wire_id*2)).get_sum_wire() if wire_id != self.N-1 else ConstantWireValue1() for wire_id in range(self.N)]
        previous_carries = [self.get_previous_component((self.N*2)-1-(wire_id*2)).get_carry_wire() for wire_id in range(self.N)]

        # Obtain proper adder name with its bit width (columns bit pairs minus the first alone bit)
        adder_name = unsigned_adder_class_name(a=a, b=b).prefix + str(self.N)
        adder_a = Bus(prefix=f"a", wires_list=previous_sums)
        adder_b = Bus(prefix=f"b", wires_list=previous_carries)

        final_adder = unsigned_adder_class_name(a=adder_a, b=adder_b, prefix=self.prefix, name=adder_name, inner_component=True, **kwargs)
        self.add_component(final_adder)
        self.out.connect_bus(connecting_bus=final_adder.out, end_connection_pos=final_adder.out.N-1, offset=(-self.N))



from ariths_gen.wire_components import (
    ConstantWireValue0,
    Bus
)
from ariths_gen.core.arithmetic_circuits import (
    GeneralCircuit
)
from ariths_gen.one_bit_circuits.one_bit_components import (
    PGSumLogic,
    GreyCell,
    BlackCell
)
from ariths_gen.one_bit_circuits.logic_gates import (
    XorGate
)
import math


class UnsignedKoggeStoneAdder(GeneralCircuit):
    """Class representing unsigned Kogge-Stone adder (using valency-2 logic gates).

    The Kogge-Stone adder belongs to a type of tree (parallel-prefix) adders.
    Tree adder structure consists of three parts of logic: 1) PG logic generation, 2) Parallel PG logic computation, 3) Final sum and cout computation
    The main difference between each tree adder lies in the implementation of the part 2).

    Along with Brent-Kung and Sklansky it is considered to be a fundamental tree adder architecture.

    Kogge-Stone achieves log2(N) stages and fanout of 2 at each stage inside 2). The circuit contains more PG cells
    than the other tree adder architectures.

    Main building components are GreyCells and BlackCells that appropriately encapsulate the essential logic used for PG computation.
    For further circuit characteristics see the book CMOS VLSI Design.

    The implementation performs the 1) and 3) (sum XORs) parts using one bit three input P/G/Sum logic function blocks.
    The 2) part is then composed according to the parallel-prefix adder characteristics.

    ```
      B3 A3         B2 A2       B1 A1       B0 A0
      │  │          │  │        │  │        │  │
    ┌─▼──▼─┐      ┌─▼──▼─┐    ┌─▼──▼─┐    ┌─▼──▼─┐
    │  PG  │  C3  │  PG  │ C2 │  PG  │ C1 │  PG  │
    │  SUM │◄────┐│  SUM │◄──┐│  SUM │◄──┐│  SUM │◄──0
    │      │     ││      │   ││      │   ││      │
    └─┬──┬┬┘     │└─┬┬┬──┘   │└─┬┬┬──┘   │└─┬┬┬──┘
      │  ││G3P3S3│  │││G2P2S2│  │││G1P1S1│  │││G0P0S0
      │ ┌▼▼──────┴──▼▼▼──────┴──▼▼▼──────┴──▼▼▼──┐
      │ │            Parallel-prefix             │
      │ │               PG logic                 │
      │ └─┬───────┬──────────┬──────────┬────────┘
      │   │S3     │S2        │S1        │S0
    ┌─▼───▼───────▼──────────▼──────────▼────────┐
    │                Sum + Cout                  │
    │                   logic                    │
    └┬────┬───────┬──────────┬──────────┬────────┘
     │    │       │          │          │
     ▼    ▼       ▼          ▼          ▼
    Cout  S3      S1         S0         S0
    ```

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        prefix (str, optional): Prefix name of unsigned ksa. Defaults to "".
        name (str, optional): Name of unsigned ksa. Defaults to "u_ksa".
    """
    def __init__(self, a: Bus, b: Bus, prefix: str = "", name: str = "u_ksa", **kwargs):
        self.N = max(a.N, b.N)
        super().__init__(inputs=[a, b], prefix=prefix, name=name, out_N=self.N+1, **kwargs)

        # Bus sign extension in case buses have different lengths
        self.a.bus_extend(N=self.N, prefix=a.prefix)
        self.b.bus_extend(N=self.N, prefix=b.prefix)
        cin = ConstantWireValue0()

        # Lists of list containing all propagate/generate wires
        self.propagate_sig = []
        self.generate_sig = []
        # Cin0 used as a first generate wire for obtaining next carry bits
        self.generate_sig.append([cin])

        # For each bit pair proceed with three stages of PPAs (Generate PG signals, Prefix computation (PG logic), Computation and connection of outputs)
        # NOTE In the implementation below, both the first and third stages are handled by the PG FAs
        for i_wire in range(self.N):
            # 1st + 3rd stage: Generate PG signals + Computation and connection of outputs
            self.add_component(PGSumLogic(self.a.get_wire(i_wire), self.b.get_wire(i_wire), self.generate_sig[i_wire][-1], prefix=self.prefix+"_pg_sum"+str(self.get_instance_num(cls=PGSumLogic))))
            self.generate_sig.append([self.get_previous_component().get_generate_wire()])
            self.propagate_sig.append([self.get_previous_component().get_propagate_wire()])

            self.out.connect(i_wire, self.get_previous_component().get_sum_wire())
            if i_wire == self.N-1:
                self.add_component(GreyCell(self.generate_sig[i_wire+1][-1], self.propagate_sig[i_wire][0], self.generate_sig[i_wire][-1], prefix=self.prefix+"_gc"+str(self.get_instance_num(cls=GreyCell))))
                self.out.connect(self.N, self.get_previous_component().get_generate_wire())
            # 2nd stage: Prefix Computation (PG logic)
            # For all bit indexes expect for the last one, proceed with the parralel prefix PG logic
            else:
                index_stages = math.ceil(math.log(i_wire+2, 2))  # +1 because indexes start from 0 and additional +1 because the first generated carry is actually the second carry after cin (the previous cin has 0 stages)
                for stage in range(index_stages):
                    # Grey cell
                    if stage == index_stages-1:
                        if stage == 0:  # Bit index with only one stage
                            self.add_component(GreyCell(self.generate_sig[i_wire+1][0], self.propagate_sig[i_wire][0], self.generate_sig[i_wire][0], prefix=self.prefix+"_gc"+str(self.get_instance_num(cls=GreyCell))))
                        else:  # Bit index contains multiple stages, GC is its last stage
                            self.add_component(GreyCell(self.get_previous_component().get_generate_wire(), self.get_previous_component().get_propagate_wire(), self.generate_sig[(i_wire+1)-(2**stage)][-1], prefix=self.prefix+"_gc"+str(self.get_instance_num(cls=GreyCell))))
                    # Black cell
                    else:  # If bit index contains more than one stage, every stage except for the last one contains a Black cells
                        self.add_component(BlackCell(self.generate_sig[i_wire+1][-1], self.propagate_sig[i_wire][-1], self.generate_sig[(i_wire+1)-(2**stage)][stage], self.propagate_sig[i_wire-(2**stage)][stage], prefix=self.prefix+"_bc"+str(self.get_instance_num(cls=BlackCell))))
                        self.propagate_sig[i_wire].append(self.get_previous_component().get_propagate_wire())
                    self.generate_sig[i_wire+1].append(self.get_previous_component().get_generate_wire())


class SignedKoggeStoneAdder(UnsignedKoggeStoneAdder, GeneralCircuit):
    """Class representing signed Kogge-Stone adder (using valency-2 logic gates).

    The Kogge-Stone adder belongs to a type of tree (parallel-prefix) adders.
    Tree adder structure consists of three parts of logic: 1) PG logic generation, 2) Parallel PG logic computation, 3) Final sum and cout computation
    The main difference between each tree adder lies in the implementation of the part 2).

    Along with Brent-Kung and Sklansky it is considered to be a fundamental tree adder architecture.

    Kogge-Stone achieves log(N,2) stages and fanout of 2 at each stage inside 2). The circuit contains more PG cells
    than the other tree adder architectures.

    Main building components are GreyCells and BlackCells that appropriately encapsulate the essential logic used for PG computation.
    For further circuit characteristics see the book CMOS VLSI Design.

    The implementation performs the 1) and 3) (sum XORs) parts using one bit three input P/G/Sum logic function blocks.
    The 2) part is then composed according to the parallel-prefix adder characteristics.
    At last XOR gates are used to ensure proper sign extension.

    ```
      B3 A3         B2 A2       B1 A1       B0 A0
      │  │          │  │        │  │        │  │
    ┌─▼──▼─┐      ┌─▼──▼─┐    ┌─▼──▼─┐    ┌─▼──▼─┐
    │  PG  │  C3  │  PG  │ C2 │  PG  │ C1 │  PG  │
    │  SUM │◄────┐│  SUM │◄──┐│  SUM │◄──┐│  SUM │◄──0
    │      │     ││      │   ││      │   ││      │
    └─┬──┬┬┘     │└─┬┬┬──┘   │└─┬┬┬──┘   │└─┬┬┬──┘
      │  ││G3P3S3│  │││G2P2S2│  │││G1P1S1│  │││G0P0S0
      │ ┌▼▼──────┴──▼▼▼──────┴──▼▼▼──────┴──▼▼▼──┐
      │ │            Parallel-prefix             │
      │ │               PG logic                 │
      │ └─┬───────┬──────────┬──────────┬────────┘
      │   │S3     │S2        │S1        │S0
    ┌─▼───▼───────▼──────────▼──────────▼────────┐
    │                Sum + Cout                  │
    │           with sign extension              │
    └┬────┬───────┬──────────┬──────────┬────────┘
     │    │       │          │          │
     ▼    ▼       ▼          ▼          ▼
    Cout  S3      S1         S0         S0
    ```

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        prefix (str, optional): Prefix name of signed ksa. Defaults to "".
        name (str, optional): Name of signed ksa. Defaults to "s_ksa".
    """
    def __init__(self, a: Bus, b: Bus, prefix: str = "", name: str = "s_ksa", **kwargs):
        # Signed bus length extension
        N = max(a.N, b.N)
        a.bus_extend(N=N, prefix=a.prefix, desired_extension_wire=a.get_wire(a.N-1))
        b.bus_extend(N=N, prefix=b.prefix, desired_extension_wire=b.get_wire(b.N-1))
        
        super().__init__(a=a, b=b, prefix=prefix, name=name, signed=True, **kwargs)

        # Additional XOR gates to ensure correct sign extension in case of sign addition
        self.add_component(XorGate(self.a.get_wire(self.N-1), self.b.get_wire(self.N-1), prefix=self.prefix+"_xor"+str(self.get_instance_num(cls=XorGate)), parent_component=self))
        self.add_component(XorGate(self.get_previous_component().out, self.get_previous_component(2).get_generate_wire(), prefix=self.prefix+"_xor"+str(self.get_instance_num(cls=XorGate)), parent_component=self))
        self.out.connect(self.N, self.get_previous_component().out)



from ariths_gen.wire_components.wires import Wire
from ariths_gen.wire_components.buses import Bus
import math


# Multiple-input #
class MultipleInputLogicGate():
    """Class representing multiple input logic gate internally composed of corresponding two input logic gates.

    Gates are cascaded in a tree like structure to solve fan-in issue concerning multi-input logic gates.
    Tree like structure increases propagation delay logarithmically with number of inputs.

    ```
       ┌──────┐
    ──►│ FUNC │
       │      ├──┐  ┌──────┐
    ──►│      │  ├─►│ FUNC │
       └──────┘  │  │      ├─►
       ┌──────┐  ├─►│      │
    ──►│ FUNC │  │  └──────┘
       │      ├──┘
    ──►│      │
       └──────┘
       ...
    ```

    Description of the __init__ method.

    Args:
        a (Bus): Bus containing input wires to form inner composite two input logic gates.
        two_input_gate_cls (type): Specific two input logic gate class type that specifies what type of multiple input logic gate the instance represents.
        parent_component (object) Object of upper component of which this logic gate is a subcomponent (used for adding composite two input gates into `parent_component`'s list of subcomponents).
        prefix (str, optional): Prefix used to name inner composite logic gates. Defaults to "".
    """
    def __init__(self, a: Bus, two_input_gate_cls, parent_component: object, prefix: str = ""):
        i = 0
        while a.N != 1:
            N = math.floor(a.N/2)
            out_wires = []
            # Creation of composite two input logic gates from bus `a`'s bit pairs and addition of generated blocks outputs for next iteration
            for bus_index in range(0, N):
                gate = two_input_gate_cls(a=a.get_wire(bus_index), b=a.get_wire(bus_index+N), prefix=prefix+ "_" + str(i) + "_" + str(parent_component.get_instance_num(cls=two_input_gate_cls, count_disabled_gates=False)), parent_component=parent_component)
                parent_component.add_component(gate)
                out_wires.append(gate.out)
                i += 1

            # In case bus `a` has odd number of wires
            if a.N % 2 != 0:
                out_wires.append(a.get_wire(-1))
                N += 1

            # Update bus for next iteration until it contains only one output wire
            a = Bus(prefix=a.prefix, N=N, wires_list=out_wires)
        self.out = a.get_wire(0)


# Two-input #
class TwoInputLogicGate():
    """Class representing two input logic gates.

    ```
       ┌──────┐
    ──►│ FUNC │
       │      ├─►
    ──►│      │
       └──────┘
    ```

    Description of the __init__ method.

    Args:
        a (Wire): First input wire.
        b (Wire): Second input wire.
        prefix (str, optional): Prefix name of logic gate. Defaults to "gate".
        outid (int, optional): Index of output wire. Defaults to 0.
        parent_component (object, optional) Object of upper component of which logic gate is a subcomponent. Defaults to None.
    """
    def __init__(self, a: Wire, b: Wire, prefix: str = "gate", outid: int = 0, parent_component: object = None):
        self.a = a
        self.b = b
        self.prefix = prefix
        self.outid = outid
        # To toggle whether logic gate function ought to be generated or to be replaced
        # by a constant wire for the sake of more optimized circuit generation
        self.disable_generation = False
        # Obtaining the caller object to gain access into its `components` list Used for adding NOT gates as a replacement for two input logic gates with constant input (optimalization)
        # Also used to obtain caller object's `prefix` name for proper wire names generation of flat/hier representations
        self.parent_component = parent_component

    """ C CODE GENERATION """
    # FLAT C #
    @staticmethod
    def get_includes_c():
        """Generates necessary C library includes for output representation.

        Returns:
            str: C code library includes.
        """
        return f"#include <stdio.h>\n#include <stdint.h>\n\n"

    def get_prototype_c_flat(self):
        """Generates flat C code function header to describe corresponding two input logic gate's interface in flat C code.

        Returns:
            str: Function's name and parameters in flat C code.
        """
        return f"uint8_t {self.prefix}(uint8_t {self.a.name}, uint8_t {self.b.name})" + "{" + "\n"

    def get_function_c(self):
        """Generates C code representing corresponding two input logic gate's Boolean function using bitwise operators between its bitwise shifted inputs.

        Returns:
            str: C code description of logic gate's Boolean function (with bitwise shifted inputs).
        """
        if self.out.is_const():
            return self.out.get_wire_value_c_flat()
        return f"{self.a.get_wire_value_c_flat()} {self.operator} {self.b.get_wire_value_c_flat()}"

    def get_declaration_c_flat(self):
        """Generates C code declaration of output wire for flat representation.

        Returns:
            str: C code logic gate's output wire declaration.
        """
        # No gate output wire is generated if one of the inputs is a wire with constant value.
        # I.e. either the constant or the second input wire is propagated to the output for the corresponding logic gate's logic function.
        if self.disable_generation:
            return ""
        else:
            return f"{self.out.get_declaration_c()}"

    def get_assign_c_flat(self):
        """Generates C code for invocation of logical functions and subsequently provides assignment to their output.

        Returns:
            str: C code invocation of logical function and assignment to output.
        """
        # No gate logic is generated if one of the inputs is a wire with constant value.
        # I.e. either the constant or the second input wire is propagated to the output for the corresponding logic gate's logic function.
        if self.disable_generation:
            return ""
        else:
            return f"  {self.out.prefix} = {self.get_function_c()};\n"

    # Generating flat C code representation of the logic gate itself
    # (I.e. not as a component of bigger circuit)
    def get_c_code(self, file_object):
        """Generates flat C code representation of corresponding logic gate itself.

        Args:
            file_object (TextIOWrapper): Destination file object where circuit's representation will be written to.
        """
        file_object.write(self.get_includes_c())
        file_object.write(self.get_prototype_c_flat())
        file_object.write("  return "+(self.get_function_c())+";\n}")

    # HIERARCHICAL C #
    def get_prototype_c_hier(self):
        """Generates hierarchical C code function header to describe corresponding two input logic gate's interface in hierarchical C code.

        Returns:
            str: Function's name and parameters in hierarchical C code.
        """
        return f"uint8_t {self.gate_type}(uint8_t {self.a.name}, uint8_t {self.b.name})" + "{" + "\n"

    def get_function_block_c(self):
        """Generates C code representation of corresponding logic gate used as a function block in hierarchical circuit description.

        Returns:
            str: C code of logic gate's function block description.
        """
        gate_block = type(self)(a=Wire(name="a"), b=Wire(name="b"))
        return f"{gate_block.get_prototype_c_hier()}" + \
               f"  return "+(gate_block.get_function_c())+";\n}\n\n"

    def get_declaration_c_hier(self):
        """Generates C code declaration of output wire for hierarchical representation.

        Returns:
            str: C code logic gate's output wire declaration.
        """
        # No gate output wire is generated if one of the inputs is a wire with constant value.
        # I.e. either the constant or the second input wire is propagated to the output for the corresponding logic gate's logic function.
        if self.disable_generation:
            return ""
        else:
            return f"{self.out.get_declaration_c()}"

    def get_gate_invocation_c(self):
        """Generates C code invocation of corresponding logic gate's generated function block.

        Returns:
            str: C code of logic gate's function block invocation.
        """
        # No function block is generated if one of the inputs is a wire with constant value.
        # I.e. either the constant or the second input wire is propagated to the output for the corresponding logic gate's logic function.
        if self.disable_generation:
            return ""
        else:
            return f"  {self.out.name} = {self.gate_type}({self.a.get_wire_value_c_hier()}, {self.b.get_wire_value_c_hier()});\n"

    """ PYTHON CODE GENERATION """
    # FLAT PYTHON #
    def get_assign_python_flat(self):
        """Generates Python code for invocation of logical functions and subsequently provides assignment to their output.

        Returns:
            str: Python code invocation of logical function and assignment to output.
        """
        # No gate logic is generated if one of the inputs is a wire with constant value.
        # I.e. either the constant or the second input wire is propagated to the output for the corresponding logic gate's logic function.
        if self.disable_generation:
            #return f"  {self.out.prefix} = {self.get_function_c()} # DD {self.prefix} \n"

            return ""
        else:
            return f"  {self.out.prefix} = {self.get_function_c()}\n"

    """ VERILOG CODE GENERATION """
    # FLAT VERILOG #
    def get_prototype_v_flat(self):
        """Generates flat Verilog module header to describe corresponding two input logic gate's interface in flat Verilog.

        Returns:
            str: Module's name and parameters in flat Verilog.
        """
        return f"module {self.prefix}(input {self.a.name}, input {self.b.name}" + \
               "".join([f", output {self.out.name}" if self.disable_generation is False else f", output {self.out.name}_out" for _ in range(1)]) + ");\n"

    def get_output_v_flat(self):
        """Generates flat Verilog module's output wire assignment used for self logic gate circuit generation.

        Returns:
            str: Module's output wire assignment in flat Verilog.
        """
        return "".join([f"  assign {self.out.name} = {self.get_function_v()};\n" if self.disable_generation is False else f"  assign {self.out.name}_out = {self.get_function_v()};\n" for _ in range(1)])

    def get_declaration_v_flat(self):
        """Generates Verilog code declaration of output wire for flat representation.

        Returns:
            str: Verilog code logic gate's output wire declaration.
        """
        # No gate output wire is generated if one of the inputs is a wire with constant value.
        # I.e. either the constant or the second input wire is propagated to the output for the corresponding logic gate's logic function.
        if self.disable_generation:
            return ""
        else:
            return f"{self.out.get_declaration_v_flat()}"

    def get_function_v(self):
        """Generates Verilog code representing corresponding two input logic gate's Boolean function using bitwise operators between its inputs.

        Returns:
            str: Verilog description of logic gate's Boolean function.
        """
        return f"{self.a.get_wire_value_v_flat()} {self.operator} {self.b.get_wire_value_v_flat()}"

    def get_assign_v_flat(self):
        """Generates Verilog code for invocation of logical functions and subsequently provides assignment to their output.

        Returns:
            str: Verilog code invocation of logical function and assignment to output.
        """
        # No gate logic is generated if one of the inputs is a wire with constant value.
        # I.e. either the constant or the second input wire is propagated to the output for the corresponding logic gate's logic function.
        if self.disable_generation:
            return ""
        else:
            return f"  assign {self.out.prefix} = {self.get_function_v()};\n"

    # Generating flat Verilog code representation of the logic gate itself
    # (I.e. not as a component of bigger circuit)
    def get_v_code(self, file_object):
        """Generates flat Verilog code representation of corresponding logic gate itself.

        Args:
            file_object (TextIOWrapper): Destination file object where circuit's representation will be written to.
        """
        file_object.write(self.get_prototype_v_flat())
        file_object.write(self.get_output_v_flat())
        file_object.write(f"endmodule")

    # HIERARCHICAL VERILOG #
    def get_prototype_v_hier(self):
        """Generates hierarchical Verilog module header to describe corresponding two input logic gate's interface in hierarchical Verilog.

        Returns:
            str: Module's name and parameters in hierarchical Verilog.
        """
        return f"module {self.gate_type}(input {self.a.name}, input {self.b.name}, output {self.out.name});\n"

    def get_function_block_v(self):
        """Generates Verilog code representation of corresponding logic gate used as function block in hierarchical circuit description.

        Returns:
            str: Verilog logic gate's function block description.
        """
        gate_block = type(self)(a=Wire(name="a"), b=Wire(name="b"), prefix="out")
        return f"{gate_block.get_prototype_v_hier()}" + \
               f"  assign {gate_block.out.name} = {gate_block.get_function_v()};\n" + \
               f"endmodule\n\n"

    def get_declaration_v_hier(self):
        """Generates Verilog code declaration of output wire for hierarchical representation.

        Returns:
            str: Verilog code logic gate's output wire declaration.
        """
        # No gate output wire is generated if one of the inputs is a wire with constant value.
        # I.e. either the constant or the second input wire is propagated to the output for the corresponding logic gate's logic function.
        if self.disable_generation:
            return ""
        else:
            return f"{self.out.get_declaration_v_hier()}"

    def get_gate_invocation_v(self):
        """Generates Verilog code invocation of corresponding logic gate's generated function block.

        Returns:
            str: Verilog code logic gate's function block invocation.
        """
        # No function block is generated if one of the inputs is a wire with constant value.
        # I.e. either the constant or the second input wire is propagated to the output for the corresponding logic gate's logic function.
        if self.disable_generation:
            return ""
        else:
            gate_block = self.__class__(a=Wire(name="a"), b=Wire(name="b"), prefix="out")
            return f"  {self.gate_type} {self.gate_type}_{self.out.prefix}(.{gate_block.a.prefix}({self.a.get_wire_value_v_hier()}), .{gate_block.b.prefix}({self.b.get_wire_value_v_hier()}), .{gate_block.out.prefix}({self.out.prefix}));\n"

    """ BLIF CODE GENERATION """
    # FLAT BLIF #
    def get_prototype_blif_flat(self):
        """Generates flat Blif model header to describe corresponding logic gate's interface in flat Blif.

        Returns:
            str: Model's name in flat Blif code.
        """
        return f".model {self.prefix}\n"

    def get_declaration_blif(self):
        """Generates Blif code declaration of two input logic gate's wires.

        Returns:
            str: Blif logic gate's wires declaration.
        """
        return f".inputs {self.a.get_wire_declaration_blif()}{self.b.get_wire_declaration_blif()}\n" + \
               f".outputs" + \
               "".join([f" {self.out.name}\n" if self.disable_generation is False else f" {self.out.name}_out\n" for _ in range(1)]) + \
               f".names vdd\n1\n" + \
               f".names gnd\n0\n"

    def get_function_blif_flat(self, top_modul: bool = False):
        """Generates Blif code representing corresponding two input logic gate's Boolean function between its inputs.

        Invokes corresponding logic gate's `get_function_blif` method for its individual description of logic function.

        Args:
            top_modul (bool, optional): Specifies whether the described circuit has logic gate as its top modul component (used for self logic gate generation). Defaults to False.

        Returns:
            str: Blif description of logic gate's Boolean function.
        """
        if top_modul is True:
            return f"{self.get_function_blif()}"
        # No function block is generated if one of the inputs is a wire with constant value.
        # I.e. either the constant or the second input wire is propagated to the output for the corresponding logic gate's logic function.
        elif self.disable_generation:
            return ""
        else:
            return f"{self.get_function_blif()}"

    # Generating flat BLIF code representation of the logic gate itself
    # (I.e. not as a component of bigger circuit)
    def get_blif_code(self, file_object):
        """Generates flat Blif code representation of corresponding logic gate itself.

        Args:
            file_object (TextIOWrapper): Destination file object where circuit's representation will be written to.
        """
        file_object.write(self.get_prototype_blif_flat())
        file_object.write(self.get_declaration_blif())
        file_object.write(self.get_function_blif_flat(top_modul=True))
        file_object.write(f".end\n")

    # HIERARCHICAL BLIF #
    def get_prototype_blif_hier(self):
        """Generates hierarchical Blif model header to describe corresponding logic gate's interface in hierarchical Blif.

        Returns:
            str: Model's name in hierarchical Blif.
        """
        return f".model {self.gate_type}\n"

    def get_function_block_blif(self):
        """Generates Blif code representation of corresponding two input logic gate used as subcomponent in hierarchical circuit description.

        Returns:
            str: Blif logic gate subcomponent description.
        """
        gate_block = type(self)(a=Wire(name="a"), b=Wire(name="b"), prefix="out")
        return f"{gate_block.get_prototype_blif_hier()}" + \
               f"{gate_block.get_declaration_blif()}" + \
               f"{gate_block.get_function_blif()}" + \
               f".end\n"

    def get_invocation_blif_hier(self, top_modul: bool = False, *args, **kwargs):
        """Generates Blif code invocation of corresponding two input logic gate's generated subcomponent.

        Args:
            top_modul (bool, optional): Specifies whether the described circuit has logic gate as its top modul component (used for self logic gate generation). Defaults to False.

        Returns:
            str: Blif logic gate subcomponent invocation.
        """
        # No function block is generated if one of the inputs is a wire with constant value.
        # I.e. either the constant or the second input wire is propagated to the output for the corresponding logic gate's logic function.
        if self.disable_generation and top_modul is False:
            return ""
        else:
            return f".subckt {self.gate_type} a={self.a.get_wire_value_blif()} b={self.b.get_wire_value_blif()} out={self.out.get_wire_value_blif()}\n"

    """ CGP CODE GENERATION """
    # FLAT CGP #
    @staticmethod
    def get_parameters_cgp():
        """Generates CGP chromosome parameters of corresponding logic gate.

        In total seven parameters represent: total inputs, total outputs, number of rows, number of columns (gates),
        number of each gate's inputs, number of each gate's outputs, quality constant value.

        Returns:
            str: CGP chromosome parameters of described logic gate.
        """
        return "{2,1,1,1,2,1,0}"

    def get_triplet_cgp(self, a_id: int, b_id: int, out_id: int):
        """Generates logic gate triplet (first input wire, second input wire, logic gate function) using wires unique position indexes within the described circuit.

        Each triplet represents unique logic gate within the described circuit. Besides the contained input wires indexes and gate's inner logic function, an output wire
        with incremented index position is also created and remembered to be appropriately driven as an input to another logic gate or as the circuit's output.

        Constant wire with value 0 has constant index of 0.
        Constant wire with value 1 has constant index of 1.
        Other wires indexes start counting from 2 and up.

        Args:
            a_id (int): First input wire index position.
            b_id (int): Second input wire index position.
            out_id (int): The output wire index position

        Returns:
            str: Triplet describing function of corresponding two input logic gate.
        """
        return f"([{out_id}]{a_id},{b_id},{self.cgp_function})"

    @staticmethod
    def get_output_cgp(out_id: int):
        """Generates list of output wires indexes of described two input logic gate from MSB to LSB.

        Args:
            out_id (int): Output wire index position.

        Returns:
            str: List containing logic gate's output wire indexes (one in this case).
        """
        return f"({out_id})"

    def get_gate_triplet_cgp(self):
        """Generates flat CGP triplet and output representation of corresponding logic gate itself.

        Returns:
            str: Triplet and output lists describing function of corresponding two input logic gate.
        """
        if self.a.is_const() and self.b.is_const():
            a_id = self.a.cgp_const
            b_id = self.b.cgp_const
        elif self.a.is_const():
            a_id = self.a.cgp_const
            b_id = 2
        elif self.b.is_const():
            a_id = 2
            b_id = self.b.cgp_const
        else:
            a_id = 2
            b_id = 3

        if self.out.is_const():
            out_id = self.out.cgp_const
        else:
            out_id = a_id+1 if a_id > b_id else b_id+1
        return self.get_triplet_cgp(a_id=a_id, b_id=b_id, out_id=out_id) + self.get_output_cgp(out_id=out_id)

    def get_cgp_code(self, file_object):
        """Generates flat CGP chromosome representation of corresponding logic gate itself.

        Args:
            file_object (TextIOWrapper): Destination file object where circuit's representation will be written to.
        """
        file_object.write(self.get_parameters_cgp())
        file_object.write(self.get_gate_triplet_cgp())


class TwoInputInvertedLogicGate(TwoInputLogicGate):
    """Class representing two input inverted logic gates.

    ```
       ┌──────┐
    ──►│ FUNC │
       │      │O──►
    ──►│      │
       └──────┘
    ```

    Description of the __init__ method.

    Args:
        a (Wire): First input wire.
        b (Wire): Second input wire.
        prefix (str, optional): Prefix name of logic gate. Defaults to "gate".
        outid (int, optional): Index of output wire. Defaults to 0.
        parent_component (object, optional) Object of upper component of which logic gate is a subcomponent. Defaults to None.
    """
    def __init__(self, a: Wire, b: Wire, prefix: str = "gate", outid: int = 0, parent_component: object = None):
        super().__init__(a, b, prefix, outid, parent_component)

    """ C CODE GENERATION """
    # FLAT C #
    def get_function_c(self):
        """Generates C code representing corresponding negated two input logic gate's Boolean function using bitwise operators between its bitwise shifted inputs.

        Returns:
            str: C code description of negated logic gate's Boolean function (with bitwise shifted inputs).
        """
        return "~("+(super().get_function_c()) + ") & 0x01"

    """ VERILOG CODE GENERATION """
    # FLAT VERILOG #
    def get_function_v(self):
        """Generates Verilog code representing corresponding negated two input logic gate's Boolean function using bitwise operators between its inputs.

        Returns:
            str: Verilog description of negated logic gate's Boolean function.
        """
        return "~("+(super().get_function_v())+")"


# One-input #
class OneInputLogicGate(TwoInputLogicGate):
    """Class representing one input logic gates.

    ```
       ┌──────┐
       │ FUNC │
    ──►│      │───►
       │      │
       └──────┘
    ```

    Description of the __init__ method.

    Args:
        a (Wire): First input wire.
        prefix (str, optional): Prefix name of logic gate. Defaults to "gate".
        outid (int, optional): Index of output wire. Defaults to 0.
        parent_component (object, optional) Object of upper component of which logic gate is a subcomponent. Defaults to None.
    """
    def __init__(self, a: Wire, prefix: str = "gate", outid: int = 0, parent_component: object = None):
        self.a = a
        self.prefix = prefix
        self.outid = outid
        # To toggle whether logic gate function ought to be generated or to be replaced
        # by a constant wire for the sake of more optimized circuit generation
        self.disable_generation = False
        # Obtaining the caller object to gain access into its `components` list Used for adding NOT gates as a replacement for two input logic gates with constant input (optimalization)
        # Also used to obtain caller object's `prefix` name for proper wire names generation of flat/hier representations
        self.parent_component = parent_component

    """ C CODE GENERATION """
    # FLAT C #
    def get_prototype_c_flat(self):
        """Generates flat C code function header to describe corresponding one input logic gate's interface in flat C code.

        Returns:
            str: Function's name and parameter in flat C code.
        """
        return f"uint8_t {self.prefix}(uint8_t {self.a.name})" + "{" + "\n"

    def get_function_c(self):
        """Generates C code representing corresponding one input logic gate's Boolean function using bitwise operators between its bitwise shifted input.

        Returns:
            str: C code description of logic gate's Boolean function (with bitwise shifted input).
        """
        return f"{self.operator}({self.a.get_wire_value_c_flat()}) & 0x01"

    # HIERARCHICAL C #
    def get_prototype_c_hier(self):
        """Generates hierarchical C code function header to describe corresponding one input logic gate's interface in hierarchical C code.

        Returns:
            str: Function's name and parameters in hierarchical C code.
        """
        return f"uint8_t {self.gate_type}(uint8_t {self.a.name})" + "{" + "\n"

    def get_function_block_c(self):
        """Generates C code representation of corresponding logic gate used as a function block in hierarchical circuit description.

        Returns:
            str: C code of logic gate's function block description.
        """
        gate_block = type(self)(a=Wire(name="a"))
        return f"{gate_block.get_prototype_c_hier()}" + \
               f"  return "+(gate_block.get_function_c())+";\n}\n\n"

    def get_gate_invocation_c(self):
        """Generates C code invocation of corresponding logic gate's generated function block.

        Returns:
            str: C code of logic gate's function block invocation.
        """
        # No function block is generated if one of the inputs is a wire with constant value.
        # I.e. either the constant or the second input wire is propagated to the output for the corresponding logic gate's logic function.
        if self.disable_generation:
            return ""
        else:
            return f"  {self.out.name} = {self.gate_type}({self.a.get_wire_value_c_hier()});\n"

    """ VERILOG CODE GENERATION """
    # FLAT VERILOG #
    def get_prototype_v_flat(self):
        """Generates flat Verilog module header to describe corresponding one input logic gate's interface in flat Verilog.

        Returns:
            str: Module's name and parameter in flat Verilog.
        """
        return f"module {self.prefix}(input {self.a.name}" + \
               "".join([f", output {self.out.name});\n" if self.disable_generation is False else f", output {self.out.name}_out);\n" for _ in range(1)])

    def get_function_v(self):
        """Generates Verilog code representing corresponding one input logic gate's Boolean function using bitwise operators between its input.

        Returns:
            str: Verilog description of logic gate's Boolean function.
        """
        return f"{self.operator}{self.a.get_wire_value_v_flat()}"

    # HIERARCHICAL VERILOG #
    def get_prototype_v_hier(self):
        """Generates hierarchical Verilog module header to describe corresponding one input logic gate's interface in hierarchical Verilog.

        Returns:
            str: Module's name and parameter in hierarchical Verilog.
        """
        return f"module {self.gate_type}(input {self.a.name}, output {self.out.name});\n"

    def get_function_block_v(self):
        """Generates Verilog code representation of corresponding logic gate used as a function block in hierarchical circuit description.

        Returns:
            str: Verilog code of logic gate's function block description.
        """
        gate_block = type(self)(a=Wire(name="a"), prefix="out")
        return f"{gate_block.get_prototype_v_hier()}" + \
               f"  assign {gate_block.out.name} = {gate_block.get_function_v()};\n" + \
               f"endmodule\n\n"

    def get_gate_invocation_v(self):
        """Generates Verilog code invocation of corresponding logic gate's generated function block.

        Returns:
            str: Verilog code logic gate's function block invocation.
        """
        # No function block is generated if one of the inputs is a wire with constant value.
        # I.e. either the constant or the second input wire is propagated to the output for the corresponding logic gate's logic function.
        if self.disable_generation:
            return ""
        else:
            gate_block = self.__class__(a=Wire(name="a"), prefix="out")
            return f"  {self.gate_type} {self.gate_type}_{self.out.prefix}(.{gate_block.a.prefix}({self.a.get_wire_value_v_hier()}), .{gate_block.out.prefix}({self.out.prefix}));\n"

    """ BLIF CODE GENERATION """
    # FLAT BLIF #
    def get_declaration_blif(self):
        """Generates Blif code declaration of one input logic gate's wires.

        Returns:
            str: Blif logic gate's wires declaration.
        """
        return f".inputs {self.a.get_wire_declaration_blif()}\n" + \
               f".outputs" + \
               "".join([f" {self.out.name}\n" if self.disable_generation is False else f" {self.out.name}_out\n" for _ in range(1)]) + \
               f".names vdd\n1\n" + \
               f".names gnd\n0\n"

    # HIERARCHICAL BLIF #
    def get_function_block_blif(self):
        """Generates Blif code representation of corresponding one input logic gate used as subcomponent in hierarchical circuit description.

        Returns:
            str: Blif logic gate subcomponent description.
        """
        gate_block = type(self)(a=Wire(name="a"), prefix="out")
        return f"{gate_block.get_prototype_blif_hier()}" + \
               f"{gate_block.get_declaration_blif()}" + \
               f"{gate_block.get_function_blif()}" + \
               f".end\n"

    def get_invocation_blif_hier(self, top_modul: bool = False, *args, **kwargs):
        """Generates Blif code invocation of corresponding one input logic gate's generated subcomponent.

        Args:
            top_modul (bool, optional): Specifies whether the described circuit has logic gate as its top modul component (used for self logic gate generation). Defaults to False.

        Returns:
            str: Blif logic gate subcomponent invocation.
        """
        # No function block is generated if one of the inputs is a wire with constant value.
        # I.e. either the constant or the second input wire is propagated to the output for the corresponding logic gate's logic function.
        if self.disable_generation and top_modul is False:
            return ""
        else:
            return f".subckt {self.gate_type} a={self.a.get_wire_value_blif()} out={self.out.get_wire_value_blif()}\n"

    """ CGP CODE GENERATION """
    # FLAT CGP #
    def get_triplet_cgp(self, a_id: int, out_id: int):
        """Generates logic gate triplet (first input wire, second input wire, logic gate function) using wires unique position indexes within the described circuit.

        Each triplet represents unique logic gate within the described circuit. In this case of one input logic gate, the same input wire index is driven to both inputs.
        Besides the contained input wires indexes and gate's inner logic function, an output wire with incremented index position is also created and remembered to be
        appropriately driven as an input to another logic gate or as the circuit's output.

        Constant wire with value 0 has constant index of 0.
        Constant wire with value 1 has constant index of 1.
        Other wires indexes start counting from 2 and up.

        Args:
            a_id (int): First (used also as the second) input wire index position.
            out_id (int): Outpu wire index position

        Returns:
            str: Triplet describing function of corresponding one input logic gate.
        """
        return f"([{out_id}]{a_id},{a_id},{self.cgp_function})"

    @staticmethod
    def get_output_cgp(out_id: int):
        """Generates list of output wires indexes of described one input logic gate from MSB to LSB.

        Args:
            out_id (int): Output wire index position.

        Returns:
            str: List containing logic gate's output wire indexes (one in this case).
        """
        return f"({out_id})"

    def get_gate_triplet_cgp(self):
        """Generates flat CGP triplet and output representation of corresponding logic gate itself.

        Returns:
            str: Triplet and output lists describing function of corresponding one input logic gate.
        """
        if self.a.is_const():
            a_id = self.a.cgp_const
        else:
            a_id = 2

        if self.out.is_const():
            out_id = self.out.cgp_const
        else:
            out_id = a_id+1 if a_id == 2 else 2
        return self.get_triplet_cgp(a_id=a_id, out_id=out_id) + self.get_output_cgp(out_id=out_id)



from ariths_gen.wire_components import (
    ConstantWireValue0,
    Bus
)
from ariths_gen.core.arithmetic_circuits import (
    MultiplierCircuit
)
from ariths_gen.one_bit_circuits.one_bit_components import (
    HalfAdder,
    FullAdder
)
from ariths_gen.one_bit_circuits.logic_gates import (
    AndGate,
)
from ariths_gen.multi_bit_circuits.adders import (
    UnsignedCarryLookaheadAdder
)


class UnsignedTruncatedCarrySaveMultiplier(MultiplierCircuit):
    """Class representing unsigned truncated carry save (Braun) multiplier.

    It represents an approximative version of unsigned carry save multiplier with simpler structure.
    It is created by modifying an ordinary N-bit unsigned carry save multiplier by ignoring
    (truncating) some of the partial products.

    The design promises better area and power parameters in exchange for the loss of computation precision.
    ```
                                                                   CUT=1
                                        A3B0       A2B0      A1B0    │    A0B0
                                       ┌───┐      ┌───┐      ┌───┐      ┌───┐
                                       │AND│      │AND│      │AND│   │  │AND│
                                       └───┘      └───┘      └───┘      └───┘
               ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─┬─ ─ ┴ ─ ─ ─ ─ ─ ─ CUT=1
                                     A3B1       A2B1       A1B1       A0B1
                                    ┌▼─▼┐      ┌▼─▼┐      ┌▼─▼┐ │    ┌───┐
                                    │AND│      │AND│      │AND│      │AND│
                                    └┬──┘      └┬──┘      └┬──┘ │    └───┘
                                     │          │          │      ┌────┐
                                 ┌───┘          │          │    │ │    │
                                 │            ┌─┘          │      │ HA │
                                 │            │            │    │ │    │
                                 │            │            │      └────┘
                                 │            │       ┌ ─ ─│─ ─ ┘
                       A3B2      │  A2B2      │  A1B2      │ A0B2
                      ┌▼─▼┐      │ ┌▼─▼┐      │ ┌───┐ │    │┌───┐
                      │AND│      │ │AND│      │ │AND│      ││AND│
                      └┬──┘      │ └┬──┘      │ └┬──┘ │    │└───┘
                       │        ┌▼──▼┐       ┌▼──▼┐     ┌──┼─┐
                       │        │    │       │    │   │ │  │ │
                       │      ┌─┤ HA │     ┌─┤ HA │     │FA│ │
                       │      │ │    │     │ │    │   │ │  │ │
                       │      │ └───┬┘     │ └───┬┘     └──┼─┘
                       │      │     │      │     │    │    │
                       │      │     │      │    ┌│─ ─ ┘    │
              A3B3     │  A2B3│     │  A1B3│     │ A0B3    │
             ┌▼─▼┐     │ ┌▼─▼┐│     │ ┌▼─▼┐│    ││┌───┐    │
             │AND│     │ │AND││     │ │AND││     ││AND│    │
             └┬──┘     │ └┬──┘│     │ └┬──┘│ ┌─ ┘│└───┘    │
              │       ┌▼──▼┐  │    ┌▼──▼┐  │ │ ┌─┼──┐      │
              │       │    │  │    │    │  │   │ │  │      │
              │    ┌──┤ FA │◄─┘  ┌─┤ FA │◄─┘ │ │ │FA│      │
              │    │  │    │     │ │    │      │ │  │      │
              │    │  └──┬─┘     │ └──┬─┘    │ └─┼──┘      │
              │    │     │       │    │      │   │         │
              │    │     │       │    │          │         │
              │    │     │       │    │          │         │
      0   0   │    │     │       │    │      │   │         │
      │   │   │S1  │C1   │S0     │C0  │          │         │
    ┌─▼───▼───▼────▼─────▼───────▼─┐  │      │   │         │
    │         Carry-propagate      │  │          │         │
    │              adder           │  │      │   │         │
    └─────┬────────┬─────────┬─────┘  │          │         │
          ▼        ▼         ▼        ▼      │   ▼         ▼         ▼      ▼
          P7       P6        P5       P4         P3        P2       P1=0   P0=0
    ```
    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        truncation_cut (int, optional): Specifies truncation cut level used in the truncated carry save multiplier circuit creation. Note: If equal to 0, the final circuit behaves as an ordinary carry save multiplier. Defaults to 0.
        prefix (str, optional): Prefix name of unsigned truncated carry save multiplier. Defaults to "".
        name (str, optional): Name of unsigned truncated carry save multiplier. Defaults to "u_tm".
        unsigned_adder_class_name (str, optional): Unsigned multi bit adder used for final vector merging of sums and carries. Defaults to UnsignedCarryLookaheadAdder.
    """
    def __init__(self, a: Bus, b: Bus, truncation_cut: int = 0, prefix: str = "", name: str = "u_tmcsa", unsigned_adder_class_name: str = UnsignedCarryLookaheadAdder, **kwargs):
        # NOTE: If truncation_cut is specified as 0 the final circuit is a simple carry save multiplier
        self.truncation_cut = truncation_cut

        self.N = max(a.N, b.N)
        # Cut level should be: 0 <= truncation_cut < N
        assert truncation_cut < self.N

        super().__init__(inputs=[a, b], prefix=prefix, name=name, out_N=self.N*2, **kwargs)

        # Bus sign extension in case buses have different lengths
        self.a.bus_extend(N=self.N, prefix=a.prefix)
        self.b.bus_extend(N=self.N, prefix=b.prefix)

        # Gradual generation of partial products
        for b_multiplier_index in range(self.truncation_cut, self.N):
            row_pp_ord = 0
            for a_multiplicand_index in range(self.truncation_cut, self.N):
                row_pp_ord += 1
                # AND gates generation for calculation of partial products
                obj_and = AndGate(self.a.get_wire(a_multiplicand_index), self.b.get_wire(b_multiplier_index), prefix=self.prefix+"_and"+str(a_multiplicand_index)+"_"+str(b_multiplier_index))
                self.add_component(obj_and)

                if a_multiplicand_index == self.truncation_cut:
                    self.out.connect(b_multiplier_index+self.truncation_cut, obj_and.out)

                # HAs generation for first row of adders (the second row in total, first contains just pp generation)
                if b_multiplier_index == self.truncation_cut+1 and a_multiplicand_index != self.N-1:
                    previous_product = self.get_previous_component((self.N-a_multiplicand_index-1)+(row_pp_ord*2-1)).out

                    obj_adder = HalfAdder(a=obj_and.out, b=previous_product, prefix=self.prefix+"_ha"+str(a_multiplicand_index)+"_"+str(b_multiplier_index))
                    self.add_component(obj_adder)

                    if (a_multiplicand_index == self.truncation_cut):
                        self.out.connect(b_multiplier_index+self.truncation_cut, obj_adder.get_sum_wire())

                # FAs generation for the remaining rows
                elif b_multiplier_index > self.truncation_cut+1 and a_multiplicand_index != self.N-1:
                    previous_sum = self.get_previous_component((self.N-(truncation_cut+1))*2-1).get_sum_wire() if a_multiplicand_index != self.N-2 else self.get_previous_component((self.N-(truncation_cut+1))*2).out
                    previous_carry = self.get_previous_component((self.N-truncation_cut)*2-1).get_carry_wire()

                    obj_adder = FullAdder(a=obj_and.out, b=previous_sum, c=previous_carry, prefix=self.prefix+"_fa"+str(a_multiplicand_index)+"_"+str(b_multiplier_index))
                    self.add_component(obj_adder)

                    if (a_multiplicand_index == self.truncation_cut):
                        self.out.connect(b_multiplier_index+self.truncation_cut, obj_adder.get_sum_wire())

                # PRODUCT GENERATION
                # 1 bit multiplier case
                if (a_multiplicand_index == self.truncation_cut and b_multiplier_index == self.truncation_cut) or (self.truncation_cut == self.N-1):
                    self.out.connect(a_multiplicand_index + b_multiplier_index, obj_and.out)

                    # 1 bit multiplier case
                    if a_multiplicand_index == self.N-1 and b_multiplier_index == self.N-1:
                        self.out.connect(a_multiplicand_index+b_multiplier_index+1, ConstantWireValue0())

        # Connecting the output bits generated from ommited cells to ground
        for grounded_out_index in range(0, self.truncation_cut*2):
            self.out.connect(grounded_out_index, ConstantWireValue0())

        # Final addition of remaining bits using chosen unsigned multi bit adder
        if self.truncation_cut != self.N-1:
            final_cpa_N = self.N - self.truncation_cut

            previous_sums = [self.get_previous_component((final_cpa_N-2-wire_id)*2).get_sum_wire() if wire_id < final_cpa_N-2 else self.get_previous_component().out if wire_id == final_cpa_N-2 else ConstantWireValue0() for wire_id in range(final_cpa_N)]
            previous_carries = [self.get_previous_component((final_cpa_N-2-wire_id)*2+2).get_carry_wire() if wire_id != final_cpa_N-1 else ConstantWireValue0() for wire_id in range(final_cpa_N)]

            # Obtain proper adder name with its bit width (columns bit pairs minus the first alone bit)
            adder_name = unsigned_adder_class_name(a=a, b=b).prefix + str(final_cpa_N)
            adder_a = Bus(prefix=f"a", wires_list=previous_sums)
            adder_b = Bus(prefix=f"b", wires_list=previous_carries)

            final_adder = unsigned_adder_class_name(a=adder_a, b=adder_b, prefix=self.prefix, name=adder_name, inner_component=True)
            self.add_component(final_adder)
            [self.out.connect(o, final_adder.out.get_wire(o-(self.N+self.truncation_cut)), inserted_wire_desired_index=o-(self.N+self.truncation_cut)) for o in range(self.N+self.truncation_cut, self.out.N)]



from ariths_gen.multi_bit_circuits.adders.ripple_carry_adder import (
    UnsignedRippleCarryAdder,
    SignedRippleCarryAdder
)

from ariths_gen.multi_bit_circuits.adders.pg_ripple_carry_adder import (
    UnsignedPGRippleCarryAdder,
    SignedPGRippleCarryAdder
)

from ariths_gen.multi_bit_circuits.adders.carry_lookahead_adder import (
    UnsignedCarryLookaheadAdder,
    SignedCarryLookaheadAdder
)

from ariths_gen.multi_bit_circuits.adders.carry_skip_adder import (
    UnsignedCarrySkipAdder,
    SignedCarrySkipAdder
)

from ariths_gen.multi_bit_circuits.adders.carry_save_adder import (
    CarrySaveAdderComponent,
    UnsignedCarrySaveAdder
)

from ariths_gen.multi_bit_circuits.adders.kogge_stone_adder import (
    UnsignedKoggeStoneAdder,
    SignedKoggeStoneAdder
)

from ariths_gen.multi_bit_circuits.adders.brent_kung_adder import (
    UnsignedBrentKungAdder,
    SignedBrentKungAdder
)

from ariths_gen.multi_bit_circuits.adders.sklansky_adder import (
    UnsignedSklanskyAdder,
    SignedSklanskyAdder
)

from ariths_gen.multi_bit_circuits.adders.han_carlson_adder import (
    UnsignedHanCarlsonAdder,
    SignedHanCarlsonAdder
)

from ariths_gen.multi_bit_circuits.adders.ladner_fischer_adder import (
    UnsignedLadnerFischerAdder,
    SignedLadnerFischerAdder
)

from ariths_gen.multi_bit_circuits.adders.knowles_adder import (
    UnsignedKnowlesAdder,
    SignedKnowlesAdder

)

from ariths_gen.multi_bit_circuits.adders.carry_select_adder import (
    UnsignedCarrySelectAdder,
    SignedCarrySelectAdder
)

from ariths_gen.multi_bit_circuits.adders.conditional_sum_adder import (
    UnsignedConditionalSumAdder,
    SignedConditionalSumAdder
)

from ariths_gen.multi_bit_circuits.adders.carry_increment_adder import (
    UnsignedCarryIncrementAdder,
    SignedCarryIncrementAdder
)



from .wire_components import (
    wires,
    buses
)

from .one_bit_circuits import (
    logic_gates,
    one_bit_components
)

from .multi_bit_circuits import (
    adders,
    multipliers,
    approximate_multipliers,
    dividers
)

from .tools  import (
    pdk, shuffle_circuit
)

__version__ = "1.1.4"


from ariths_gen.multi_bit_circuits.approximate_multipliers.truncated_array_multiplier import (
    UnsignedTruncatedArrayMultiplier
)

from ariths_gen.multi_bit_circuits.approximate_multipliers.broken_array_multiplier import (
    UnsignedBrokenArrayMultiplier
)

from ariths_gen.multi_bit_circuits.approximate_multipliers.broken_carry_save_multiplier import (
    UnsignedBrokenCarrySaveMultiplier
)

from ariths_gen.multi_bit_circuits.approximate_multipliers.truncated_carry_save_multiplier import (
    UnsignedTruncatedCarrySaveMultiplier
)

from ariths_gen.multi_bit_circuits.approximate_multipliers.recursive_multiplier import (
    UnsignedAccurateTwoBitMultiplier,
    UnsignedApproximateTwoBitMultiplierM1,
    UnsignedApproximateTwoBitMultiplierM2,
    UnsignedApproximateTwoBitMultiplierM3,
    UnsignedApproximateTwoBitMultiplierM4,
    UnsignedRecursiveMultiplier
)



from ariths_gen.wire_components import (
    ConstantWireValue0,
    Bus
)
from ariths_gen.core.arithmetic_circuits import (
    GeneralCircuit
)
from ariths_gen.one_bit_circuits.one_bit_components import (
    TwoOneMultiplexer,
    FullSubtractor
)
from ariths_gen.one_bit_circuits.logic_gates import (
    NotGate
)


class ArrayDivider(GeneralCircuit):
    """Class representing array divider.

    Array divider performs division between two N bit numbers and stores their
    quotient in the output bus (remainder bits are not returned).

    Design is based on series of iterative subtractions.

    ```
           0  B3     0  B2     0  B1     A3 B0
           │  │      │  │      │  │      │  │
           │  │      │  │      │  │      │  │
         ┌─▼──▼─┐  ┌─▼──▼─┐  ┌─▼──▼─┐  ┌─▼──▼─┐
         │      │  │      │  │      │  │      │
       ┌─┤  FS  │◄─┤  FS  │◄─┤  FS  │◄─┤  FS  │◄─ 0
       │ │      │  │      │  │      │  │      │
       │ └──────┘  └───┬──┘  └───┬──┘  └───┬──┘
       │               │   │0    │   │0    │   │A3
       │             ┌─▼───▼┐  ┌─▼───▼┐  ┌─▼───▼┐
       ├────────────►│2:1MUX├─►│2:1MUX├─►│2:1MUX│
       │             └───┬──┘  └───┬──┘  └───┬──┘
       │              MX0│ B3   MX1│ B2   MX2│  B1     A2 B0
       │                 │ │       │ │       │  │      │  │
       │               ┌─▼─▼──┐  ┌─▼─▼──┐  ┌─▼──▼─┐  ┌─▼──▼─┐
       │               │      │  │      │  │      │  │      │
       │             ┌─┤  FS  │◄─┤  FS  │◄─┤  FS  │◄─┤  FS  │◄─ 0
       │             │ │      │  │      │  │      │  │      │
       │             │ └──────┘  └───┬──┘  └───┬──┘  └───┬──┘
       │             │               │   │MX1  │   │MX2  │   │A2
       │             │             ┌─▼───▼┐  ┌─▼───▼┐  ┌─▼───▼┐
       │             ├────────────►│2:1MUX├─►│2:1MUX│─►│2:1MUX│
       │             │             └───┬──┘  └───┬──┘  └───┬──┘
       │             │              MX3│ B3   MX4│ B2   MX5│  B1     A1 B0
       │             │                 │ │       │ │       │  │      │  │
       │             │               ┌─▼─▼──┐  ┌─▼─▼──┐  ┌─▼──▼─┐  ┌─▼──▼─┐
       │             │               │      │  │      │  │      │  │      │
       │             │             ┌─┤  FS  │◄─┤  FS  │◄─┤  FS  │◄─┤  FS  │◄─ 0
       │             │             │ │      │  │      │  │      │  │      │
       │             │             │ └──────┘  └───┬──┘  └───┬──┘  └───┬──┘
       │             │             │               │   │MX4  │   │MX5  │   │A1
       │             │             │             ┌─▼───▼┐  ┌─▼───▼┐  ┌─▼───▼┐
       │             │             │────────────►│2:1MUX├─►│2:1MUX│─►│2:1MUX│
       │             │             │             └───┬──┘  └───┬──┘  └───┬──┘
       │             │             │              MX6│ B3   MX7│ B2   MX8│  B1     A0 B0
       │             │             │                 │ │       │ │       │  │      │  │
       │             │             │               ┌─▼─▼──┐  ┌─▼─▼──┐  ┌─▼──▼─┐  ┌─▼──▼─┐
       │             │             │               │      │  │      │  │      │  │      │
       │             │             │             ┌─┤  FS  │◄─┤  FS  │◄─┤  FS  │◄─┤  FS  │◄─ 0
       │             │             │             │ │      │  │      │  │      │  │      │
       │             │             │             │ └──────┘  └──────┘  └──────┘  └──────┘
       │             │             │             │
       │             │             │             │
    ┌──▼──┐       ┌──▼──┐       ┌──▼──┐       ┌──▼──┐
    │     │       │     │       │     │       │     │
    │ NOT │       │ NOT │       │ NOT │       │ NOT │
    │     │       │     │       │     │       │     │
    └──┬──┘       └──┬──┘       └──┬──┘       └──┬──┘
       ▼             ▼             ▼             ▼
       Q3            Q2            Q1            Q0
    ```

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        prefix (str, optional): Prefix name of array divider. Defaults to "".
        name (str, optional): Name of array divider. Defaults to "arrdiv".
    """
    def __init__(self, a: Bus, b: Bus, prefix: str = "", name: str = "arrdiv", **kwargs):
        self.N = max(a.N, b.N)
        super().__init__(inputs=[a, b], prefix=prefix, name=name, out_N=self.N, **kwargs)

        # Bus sign extension in case buses have different lengths
        self.a.bus_extend(N=self.N, prefix=a.prefix)
        self.b.bus_extend(N=self.N, prefix=b.prefix)

        # Performing series of iterative subtractions
        # Gradually shifting the divisor
        for a_index in reversed(range(self.N)):
            # Series of subtractions to obtain quotient bit
            for b_index in range(self.N):
                # First full subtractor is formed from divisor's LSB bit (b_index) and divident's MSB bit (a_index)
                if b_index == 0:
                    adder_object = FullSubtractor(a=self.a.get_wire(a_index), b=self.b.get_wire(b_index), c=ConstantWireValue0(), prefix=self.prefix+"_fs"+str(self.get_instance_num(cls=FullSubtractor)))
                elif a_index == self.N-1:
                    adder_object = FullSubtractor(a=ConstantWireValue0(), b=self.b.get_wire(b_index), c=self.get_previous_component().get_borrow_wire(), prefix=self.prefix+"_fs"+str(self.get_instance_num(cls=FullSubtractor)))
                else:
                    adder_object = FullSubtractor(a=self.get_previous_component(self.N+1).out.get_wire(), b=self.b.get_wire(b_index), c=self.get_previous_component().get_borrow_wire(), prefix=self.prefix+"_fs"+str(self.get_instance_num(cls=FullSubtractor)))
                self.add_component(adder_object)

            # Don't generate multiplexers for divison remainders
            if a_index != 0:
                for mux_index in range(self.N-1):
                    mux_object = TwoOneMultiplexer(a=self.get_previous_component(self.N).get_difference_wire(), b=self.get_previous_component(self.N).a, c=self.get_previous_component(1+mux_index).get_borrow_wire(), prefix=self.prefix+"_mux2to1"+str(self.get_instance_num(cls=TwoOneMultiplexer)))
                    self.add_component(mux_object)

            # Every borrow out obtained from each iteration of subtractions needs to be negated to represent the quotient output bit
            quotient = NotGate(a=adder_object.get_borrow_wire(), prefix=self.prefix+"_not"+str(self.get_instance_num(cls=NotGate)), parent_component=self)
            self.add_component(quotient)
            self.out.connect(a_index, quotient.out)



from ariths_gen.wire_components import (
    ConstantWireValue0,
    Bus
)
from ariths_gen.core.arithmetic_circuits import (
    MultiplierCircuit
)
from ariths_gen.one_bit_circuits.logic_gates import (
    AndGate,
    OrGate,
    XorGate,
    NotGate
)
from ariths_gen.multi_bit_circuits.adders import (
    UnsignedCarryLookaheadAdder
)
import math


class UnsignedAccurateTwoBitMultiplier(MultiplierCircuit):
    """Class representing unsigned two-bit accurate multiplier.

    ```
         A1B1   A1B0   A0B1   A0B0
         │ │    │ │    │ │    │ │
        ┌▼─▼┐  ┌▼─▼┐  ┌▼─▼┐  ┌▼─▼┐
        │AND│  │AND│  │AND│  │AND│
        └─┬─┘  └─┬─┘  └─┬─┘  └─┬─┘
          │      └──┬──┐└┬─┐   │
          │         │  │ │ │   │
          │         │  │ │ │   │
          │         │ ┌▼─▼┐│   │
          │         │ │AND││   │
          │         │ └─┬─┘│   │
          └─────┐ ┌─┼───┘┌─┘   │
                │ │ └──┐ │     │
        ┌◄─►┐  ┌▼─▼┐  ┌▼─▼┐    │
        │AND│  │XOR│  │XOR│    │
        └─┬─┘  └─┬─┘  └─┬─┘    │
          │      │      │      │
          ▼      ▼      ▼      ▼
          O3     O2     O1     O0
    ```

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        prefix (str, optional): Prefix name of unsigned two-bit accurate multiplier. Defaults to "".
        name (str, optional): Name of unsigned two-bit accurate multiplier. Defaults to "u_2bit_accm".
    """
    def __init__(self, a: Bus, b: Bus, prefix: str = "", name: str = "u_2bit_accm", **kwargs):
        self.N = max(a.N, b.N)
        assert self.N == 2
        super().__init__(inputs=[a, b], prefix=prefix, name=name, out_N=self.N*2, **kwargs)

        # Bus sign extension in case buses have different lengths
        self.a.bus_extend(N=self.N, prefix=a.prefix)
        self.b.bus_extend(N=self.N, prefix=b.prefix)

        and_obj1 = AndGate(self.a.get_wire(0), self.b.get_wire(0), prefix=self.prefix+"_and0")
        and_obj2 = AndGate(self.a.get_wire(0), self.b.get_wire(1), prefix=self.prefix+"_and1")
        and_obj3 = AndGate(self.a.get_wire(1), self.b.get_wire(0), prefix=self.prefix+"_and2")
        and_obj4 = AndGate(self.a.get_wire(1), self.b.get_wire(1), prefix=self.prefix+"_and3")

        xor_obj1 = XorGate(and_obj2.out, and_obj3.out, prefix=self.prefix+"_xor0")
        and_obj5 = AndGate(and_obj2.out, and_obj3.out, prefix=self.prefix+"_and4")
        xor_obj2 = XorGate(and_obj5.out, and_obj4.out, prefix=self.prefix+"_xor1")
        and_obj6 = AndGate(and_obj5.out, and_obj4.out, prefix=self.prefix+"_and5")
        [self.add_component(obj) for obj in [and_obj1, and_obj2, and_obj3, and_obj4, xor_obj1, and_obj5, xor_obj2, and_obj6]]

        self.out.connect(0, and_obj1.out)
        self.out.connect(1, xor_obj1.out)
        self.out.connect(2, xor_obj2.out)
        self.out.connect(3, and_obj6.out)


class SignedAccurateTwoBitMultiplier(MultiplierCircuit):
    """Class representing signed two-bit accurate multiplier.

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        prefix (str, optional): Prefix name of signed two-bit accurate multiplier. Defaults to "".
        name (str, optional): Name of signed two-bit accurate multiplier. Defaults to "s_2bit_accm".
    """
    def __init__(self, a: Bus, b: Bus, prefix: str = "", name: str = "s_2bit_accm", **kwargs):
        self.N = max(a.N, b.N)
        raise NotImplementedError("SignedAccurateTwoBitMultiplier is not implemented yet.")


class UnsignedApproximateTwoBitMultiplierM1(MultiplierCircuit):
    """Class representing unsigned two-bit approximate multiplier variant M1.

    M1 ax variant defined here: https://ieeexplore.ieee.org/document/8727537

    ```
         A1B1   A1B0   A0B1   A0B0
         │ │    │ │    │ │    │ │
        ┌▼─▼┐  ┌▼─▼┐  ┌▼─▼┐  ┌▼─▼┐
        │AND│  │AND│  │AND│  │AND│
        └─┬─┘  └─┬─┘  └─┬─┘  └─┬─┘
          │      │      └┐     │
          │      └─────┐ │     │
          └──────┐    ┌▼─▼┐    │
                 │    │ OR│    │
                 │    └─┬─┘    │
                 │      │      │
          ▼      ▼      ▼      ▼
        O3=0     O2     O1     O0
    ```

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        prefix (str, optional): Prefix name of unsigned two-bit approximate multiplier m1. Defaults to "".
        name (str, optional): Name of unsigned two-bit approximate multiplier m1. Defaults to "u_2bit_axm1".
    """
    def __init__(self, a: Bus, b: Bus, prefix: str = "", name: str = "u_2bit_axm1", **kwargs):
        self.N = max(a.N, b.N)
        assert self.N == 2
        super().__init__(inputs=[a, b], prefix=prefix, name=name, out_N=self.N*2, **kwargs)

        # Bus sign extension in case buses have different lengths
        self.a.bus_extend(N=self.N, prefix=a.prefix)
        self.b.bus_extend(N=self.N, prefix=b.prefix)

        and_obj1 = AndGate(self.a.get_wire(0), self.b.get_wire(0), prefix=self.prefix+"_and0")
        and_obj2 = AndGate(self.a.get_wire(0), self.b.get_wire(1), prefix=self.prefix+"_and1")
        and_obj3 = AndGate(self.a.get_wire(1), self.b.get_wire(0), prefix=self.prefix+"_and2")
        and_obj4 = AndGate(self.a.get_wire(1), self.b.get_wire(1), prefix=self.prefix+"_and3")

        or_obj1 = OrGate(and_obj2.out, and_obj3.out, prefix=self.prefix+"_or0")
        [self.add_component(obj) for obj in [and_obj1, and_obj2, and_obj3, and_obj4, or_obj1]]

        self.out.connect(0, and_obj1.out)
        self.out.connect(1, or_obj1.out)
        self.out.connect(2, and_obj4.out)
        self.out.connect(3, ConstantWireValue0())


class SignedApproximateTwoBitMultiplierM1(MultiplierCircuit):
    """Class representing signed two-bit approximate multiplier variant M1.

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        prefix (str, optional): Prefix name of signed two-bit approximate multiplier m1. Defaults to "".
        name (str, optional): Name of signed two-bit approximate multiplier m1. Defaults to "s_2bit_axm1".
    """
    def __init__(self, a: Bus, b: Bus, prefix: str = "", name: str = "s_2bit_axm1", **kwargs):
        raise NotImplementedError("SignedApproximateTwoBitMultiplierM1 is not implemented yet.")


class UnsignedApproximateTwoBitMultiplierM2(MultiplierCircuit):
    """Class representing unsigned two-bit approximate multiplier variant M2.

    M2 ax variant defined here: https://ieeexplore.ieee.org/document/8727537

    ```
         A1B1          A1B0   A0B1
         │ │           │ │    │ │
        ┌▼─▼┐         ┌▼─▼┐  ┌▼─▼┐
        │AND│         │AND│  │AND│
        └─┬─┘         └─┬─┘  └─┬─┘
          │             │┌─────┴┐
          │            ┌┴┼────┐ │
          │            │ │    │ │
          │           ┌▼─▼┐  ┌▼─▼┐
          │           │XOR│  │AND│
          │           └─┬─┘  └─┬─┘
          └─────┐ ┌─────┼──────┤
                │ │     │      │
               ┌▼─▼┐    │      │
               │XOR│    │      │
               └─┬─┘    │      │
          ┌──────┼──────┼──────┤
          │      │      │      │
          ▼      ▼      ▼      ▼
          O3     O2     O1     O0
    ```

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        prefix (str, optional): Prefix name of unsigned two-bit approximate multiplier m1. Defaults to "".
        name (str, optional): Name of unsigned two-bit approximate multiplier m1. Defaults to "u_2bit_axm1".
    """
    def __init__(self, a: Bus, b: Bus, prefix: str = "", name: str = "u_2bit_axm1", **kwargs):
        self.N = max(a.N, b.N)
        assert self.N == 2
        super().__init__(inputs=[a, b], prefix=prefix, name=name, out_N=self.N*2, **kwargs)

        # Bus sign extension in case buses have different lengths
        self.a.bus_extend(N=self.N, prefix=a.prefix)
        self.b.bus_extend(N=self.N, prefix=b.prefix)

        and_obj1 = AndGate(self.a.get_wire(0), self.b.get_wire(1), prefix=self.prefix+"_and0")
        and_obj2 = AndGate(self.a.get_wire(1), self.b.get_wire(0), prefix=self.prefix+"_and1")
        and_obj3 = AndGate(self.a.get_wire(1), self.b.get_wire(1), prefix=self.prefix+"_and2")

        and_obj4 = AndGate(and_obj1.out, and_obj2.out, prefix=self.prefix+"_and3")
        xor_obj1 = XorGate(and_obj1.out, and_obj2.out, prefix=self.prefix+"_xor0")

        xor_obj2 = XorGate(and_obj4.out, and_obj3.out, prefix=self.prefix+"_xor1")
        [self.add_component(obj) for obj in [and_obj1, and_obj2, and_obj3, and_obj4, xor_obj1, xor_obj2]]

        self.out.connect(0, and_obj4.out)
        self.out.connect(1, xor_obj1.out)
        self.out.connect(2, xor_obj2.out)
        self.out.connect(3, and_obj4.out)


class SignedApproximateTwoBitMultiplierM2(MultiplierCircuit):
    """Class representing signed two-bit approximate multiplier variant M2.

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        prefix (str, optional): Prefix name of signed two-bit approximate multiplier m2. Defaults to "".
        name (str, optional): Name of signed two-bit approximate multiplier m2. Defaults to "s_2bit_axm2".
    """
    def __init__(self, a: Bus, b: Bus, prefix: str = "", name: str = "s_2bit_axm2", **kwargs):
        raise NotImplementedError("SignedApproximateTwoBitMultiplierM2 is not implemented yet.")


class UnsignedApproximateTwoBitMultiplierM3(MultiplierCircuit):
    """Class representing unsigned two-bit approximate multiplier variant M3.

    M3 ax variant defined here: https://ieeexplore.ieee.org/document/8727537

    ```
         A1B1   A1B0   A0B1   A0B0
         │ │    │ │    │ │    │ │
        ┌▼─▼┐  ┌▼─▼┐  ┌▼─▼┐  ┌▼─▼┐
        │AND│  │AND│  │AND│  │AND│
        └─┬─┘  └─┬─┘  └─┬─┘  └─┬─┘
          │      │      └┐     │
          │      │       │     │
          │      └─────┐ │     │
          │           ┌▼─▼┐    │
          │           │ OR│    │
          │           └─┬─┘    │
          │┌─────┬──────┼──────┤
          ││     │      │      │
          ││   ┌─▼─┐    │      │
          ││   │NOT│    │      │
          ││   └─┬─┘    │      │
         ┌┴┼────┐└┐     │      │
         │ │    │ │     │      │
        ┌▼─▼┐  ┌▼─▼┐    │      │
        │AND│  │AND│    │      │
        └─┬─┘  └─┬─┘    │      │
          │      │      │      │
          ▼      ▼      ▼      ▼
          O3     O2     O1     O0
    ```

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        prefix (str, optional): Prefix name of unsigned two-bit approximate multiplier m3. Defaults to "".
        name (str, optional): Name of unsigned two-bit approximate multiplier m3. Defaults to "u_2bit_axm3".
    """
    def __init__(self, a: Bus, b: Bus, prefix: str = "", name: str = "u_2bit_axm3", **kwargs):
        self.N = max(a.N, b.N)
        assert self.N == 2
        super().__init__(inputs=[a, b], prefix=prefix, name=name, out_N=self.N*2, **kwargs)

        # Bus sign extension in case buses have different lengths
        self.a.bus_extend(N=self.N, prefix=a.prefix)
        self.b.bus_extend(N=self.N, prefix=b.prefix)

        and_obj1 = AndGate(self.a.get_wire(0), self.b.get_wire(0), prefix=self.prefix+"_and0")
        and_obj2 = AndGate(self.a.get_wire(0), self.b.get_wire(1), prefix=self.prefix+"_and1")
        and_obj3 = AndGate(self.a.get_wire(1), self.b.get_wire(0), prefix=self.prefix+"_and2")
        and_obj4 = AndGate(self.a.get_wire(1), self.b.get_wire(1), prefix=self.prefix+"_and3")

        or_obj1 = OrGate(and_obj2.out, and_obj3.out, prefix=self.prefix+"_xor0")
        not_obj1 = NotGate(and_obj1.out, prefix=self.prefix+"_not0")

        and_obj5 = AndGate(not_obj1.out, and_obj4.out, prefix=self.prefix+"_and4")
        and_obj6 = AndGate(and_obj1.out, and_obj4.out, prefix=self.prefix+"_and5")
        [self.add_component(obj) for obj in [and_obj1, and_obj2, and_obj3, and_obj4, or_obj1, not_obj1, and_obj5, and_obj6]]

        self.out.connect(0, and_obj1.out)
        self.out.connect(1, or_obj1.out)
        self.out.connect(2, and_obj5.out)
        self.out.connect(3, and_obj6.out)


class SignedApproximateTwoBitMultiplierM3(MultiplierCircuit):
    """Class representing signed two-bit approximate multiplier variant M3.

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        prefix (str, optional): Prefix name of signed two-bit approximate multiplier m3. Defaults to "".
        name (str, optional): Name of signed two-bit approximate multiplier m3. Defaults to "s_2bit_axm3".
    """
    def __init__(self, a: Bus, b: Bus, prefix: str = "", name: str = "s_2bit_axm3", **kwargs):
        raise NotImplementedError("SignedApproximateTwoBitMultiplierM3 is not implemented yet.")


class UnsignedApproximateTwoBitMultiplierM4(MultiplierCircuit):
    """Class representing unsigned two-bit approximate multiplier variant M4.

    M4 ax variant defined here: https://ieeexplore.ieee.org/document/8727537

    ```
         A1B1   A1B0   A0B1   A0B0
         │ │    │ │    │ │    │ │
        ┌▼─▼┐  ┌▼─▼┐  ┌▼─▼┐  ┌▼─▼┐
        │AND│  │AND│  │AND│  │AND│
        └─┬─┘  └─┬─┘  └─┬─┘  └─┬─┘
          │      │      └┐     │
          │      └─────┐ │     │
          └──────┐    ┌▼─▼┐    │
                 │    │XOR│    │
                 │    └─┬─┘    │
                 │      │      │
          ▼      ▼      ▼      ▼
        O3=0     O2     O1     O0
    ```

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        prefix (str, optional): Prefix name of unsigned two-bit approximate multiplier m4. Defaults to "".
        name (str, optional): Name of unsigned two-bit approximate multiplier m4. Defaults to "u_2bit_axm4".
    """
    def __init__(self, a: Bus, b: Bus, prefix: str = "", name: str = "u_2bit_axm4", **kwargs):
        self.N = max(a.N, b.N)
        assert self.N == 2
        super().__init__(inputs=[a, b], prefix=prefix, name=name, out_N=self.N*2, **kwargs)

        # Bus sign extension in case buses have different lengths
        self.a.bus_extend(N=self.N, prefix=a.prefix)
        self.b.bus_extend(N=self.N, prefix=b.prefix)

        and_obj1 = AndGate(self.a.get_wire(0), self.b.get_wire(0), prefix=self.prefix+"_and0")
        and_obj2 = AndGate(self.a.get_wire(0), self.b.get_wire(1), prefix=self.prefix+"_and1")
        and_obj3 = AndGate(self.a.get_wire(1), self.b.get_wire(0), prefix=self.prefix+"_and2")
        and_obj4 = AndGate(self.a.get_wire(1), self.b.get_wire(1), prefix=self.prefix+"_and3")

        xor_obj1 = XorGate(and_obj2.out, and_obj3.out, prefix=self.prefix+"_xor0")
        [self.add_component(obj) for obj in [and_obj1, and_obj2, and_obj3, and_obj4, xor_obj1]]

        self.out.connect(0, and_obj1.out)
        self.out.connect(1, xor_obj1.out)
        self.out.connect(2, and_obj4.out)
        self.out.connect(3, ConstantWireValue0())


class SignedApproximateTwoBitMultiplierM4(MultiplierCircuit):
    """Class representing signed two-bit approximate multiplier variant M4.

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        prefix (str, optional): Prefix name of signed two-bit approximate multiplier m4. Defaults to "".
        name (str, optional): Name of signed two-bit approximate multiplier m4. Defaults to "s_2bit_axm4".
    """
    def __init__(self, a: Bus, b: Bus, prefix: str = "", name: str = "s_2bit_axm4", **kwargs):
        raise NotImplementedError("SignedApproximateTwoBitMultiplierM4 is not implemented yet.")


class UnsignedRecursiveMultiplier(MultiplierCircuit):
    """Class representing unsigned recursive multiplier.

    Input bit-vector length N can be any power of two greater than 1 (e.g. 2, 4, 8, ...).

    The internal structure of the recursive multiplier is composed of subsequent two-bit submultipliers provided in the input `submultipliers` list.
    The `submultipliers` list should contain the classes of the two-bit submultipliers that will be used for instantiation. If None are provided, accurate two-bit submultipliers are assumed.

    The number of submultipliers required is equal to (N/2)² for N > 2. For N = 2, only one two-bit submultiplier is required.

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        prefix (str, optional): Prefix name of unsigned recursive multiplier. Defaults to "".
        name (str, optional): Name of unsigned recursive multiplier. Defaults to "u_rm".
        submultipliers (list, optional): List of composite two bit submultiplier classes for instantiation. If None are provided, accurate submultipliers are assumed. Defaults to None.
                        Defaults (if None) to the required number of UnsignedAccurateTwoBitMultiplier instances.
        unsigned_adder_class_name (str, optional): Unsigned multi bit adder used to obtain final sums of products. Defaults to UnsignedCarryLookaheadAdder.
    """
    def __init__(self, a: Bus, b: Bus, prefix: str = "", name: str = "u_rm", submultipliers: list = None, unsigned_adder_class_name: str = UnsignedCarryLookaheadAdder, **kwargs):
        self.N = max(a.N, b.N)
        assert self.N > 1 and self.N & (self.N-1) == 0  # assure that N is a power of two greater than 1 (So allowed N is 2, 4, 8, ..)
        super().__init__(inputs=[a, b], prefix=prefix, name=name, out_N=self.N*2, **kwargs)

        # Bus sign extension in case buses have different lengths
        self.a.bus_extend(N=self.N, prefix=a.prefix)
        self.b.bus_extend(N=self.N, prefix=b.prefix)

        if submultipliers is None:  # By default, we assume composition from accurate two bit submultipliers
            if self.N == 2:
                submultipliers = [UnsignedAccurateTwoBitMultiplier]
            else:
                submultipliers = [UnsignedAccurateTwoBitMultiplier for _ in range((self.N//2)**2)]

        assert (self.N > 2 and len(submultipliers) == (self.N//2)**2) or (self.N == 2 and len(submultipliers) == 1)

        if self.N == 2:  # Base case for just one two-bit multiplier
            # TODO add suffix in ariths_gen rework
            mult = submultipliers[0](Bus(prefix=self.prefix + "_a", wires_list=self.a.bus), Bus(prefix=self.prefix + "_b", wires_list=self.b.bus), prefix=self.prefix + "_" + str(self.get_instance_num(cls=submultipliers[0])), **kwargs)
            self.add_component(mult)
            self.out.connect_bus(mult.out)
        else:
            # Levels of construction of the recursive multiplier
            # recursive_levels = int(math.log2(self.N)-1)  # Number of recursive levels based on the power ith power of 2 (e.g. for N=8, we have 2 recursive levels)
            block_level = 1
            partial_products = []

            for m in range(len(submultipliers)):  # Iterate over all 2-bit submultipliers (equals range(0, 4**recursive_levels))
                a_bus_offset = 0
                b_bus_offset = 0
                curr_level = block_level
                curr_id = m

                # Determine the wires offsets from MSB (for input bus `a` and `b`) for the current 2-bit submultiplier
                # There is a pattern, for example for N=8, we have 16 two-bit submultipliers with offsets:
                #       Mult ID:  1  2  3  4   5  6  7  8   9 10 11 12   13 14 15 16
                # MSB a_offsets:  0  0  2  2   0  0  2  2   4  4  6  6    4  4  6  6
                # MSB b_offsets:  0  2  0  2   4  6  4  6   0  2  0  2    4  6  4  6
                while curr_level != 0:
                    # A offset
                    if curr_id // ((4**curr_level)//2) != 0:
                        a_bus_offset += 2**curr_level

                    # B offset
                    if (curr_id // ((4**curr_level)//4)) % 2 != 0:
                        b_bus_offset += 2**curr_level

                    curr_level -= 1
                    curr_id -= (4**curr_level)*((curr_id // (4**curr_level)))

                # Create 2-bit submultiplier with the corresponding input bits
                # TODO add suffix in ariths_gen rework
                submult_a_bus = Bus(prefix=f"mult{m}_a", wires_list=self.a.bus[::-1][0+a_bus_offset:0+a_bus_offset+2][::-1], N=2)
                submult_b_bus = Bus(prefix=f"mult{m}_b", wires_list=self.b.bus[::-1][0+b_bus_offset:0+b_bus_offset+2][::-1], N=2)
                submult = submultipliers[m](submult_a_bus, submult_b_bus, prefix=self.prefix + "_" + str(self.get_instance_num(cls=submultipliers[m])))
                self.add_component(submult)

                # Create wire vectors holding partial products for final summation
                pp = Bus(prefix=f"pp_{m}", N=self.out.N, wires_list=[ConstantWireValue0() for _ in range(self.out.N)])
                # [pp.connect_bus(submult.out, offset=(self.out.N-4)-(a_bus_offset+b_bus_offset))]
                [pp.connect((self.out.N-1)-(a_bus_offset+b_bus_offset)-i, submult.out[3-i], inserted_wire_desired_index=3-i) for i in range(4)]
                partial_products.append(pp)

                # Distinction of levels of blocks to properly instantiate and connect 2-bit submultipliers
                # For example, for N=8, we have 4 two-bit submultipliers in the first level, but then we have 4 times the
                # same structure (4 two-bit mults) as the base component for the second level, similarly for N=16, but
                # with additional third layer (consisting of 16 two-bit submultipliers)
                if (m+1) % (4**block_level) == 0:
                    block_level += 1  # Increase the block level

            # Create tree of partial product adders
            while len(partial_products) != 1:
                N = len(partial_products)//2
                # Creation of composite unsigned multi bit adders from set of partial product vectors and addition of generated blocks outputs for next iteration
                for bus_index in range(0, N):
                    # TODO arithsgen_rework: update check for bus declaration and assignment (if true do declare/assign again - here we would not create new bus, just assign the existing one); it should create cleaner outcode with unncecessary new bus declarations
                    adder_name = unsigned_adder_class_name(a=a, b=b).prefix + str(self.get_instance_num(cls=unsigned_adder_class_name))
                    # TODO rework the buses
                    bus_a = Bus(prefix=f"{adder_name}_a", wires_list=partial_products[bus_index].bus, out_bus=True) if partial_products[bus_index].out_bus else partial_products[bus_index]
                    bus_b = Bus(prefix=f"{adder_name}_b", wires_list=partial_products[bus_index+N].bus, out_bus=True) if partial_products[bus_index+N].out_bus else partial_products[bus_index+N]
                    adder = unsigned_adder_class_name(a=bus_a, b=bus_b, prefix=self.prefix, name=adder_name, inner_component=True, **kwargs)
                    self.add_component(adder)
                    partial_products.append(adder.out)

                # Update partial products list for next iteration until it contains only one output vector
                partial_products = partial_products[2*N:]
            # Connect the final output of the recursive multiplier
            self.out.connect_bus(partial_products[0])


class SignedRecursiveMultiplier(MultiplierCircuit):
    """Class representing signed recursive multiplier.

    TODO
    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        prefix (str, optional): Prefix name of signed recursive multiplier. Defaults to "".
        name (str, optional): Name of signed recursive multiplier. Defaults to "s_rm".
    """
    def __init__(self, a: Bus, b: Bus, prefix: str = "", name: str = "s_rm", **kwargs):
        raise NotImplementedError("SignedRecursiveMultiplier is not implemented yet.")



from ariths_gen.core.arithmetic_circuits import (
    GeneralCircuit
)

from ariths_gen.wire_components.wires import Wire


class TwoInputOneBitCircuit(GeneralCircuit):
    """Class represents a general two input one bit circuit and implements their generation to various representations. It is derived from `ArithmeticCircuit` class.

    Description of the __init__ method.

    Args:
        a (Wire): First input wire.
        b (Wire): Second input wire.
        prefix (str, optional): Prefix name of circuit. Defaults to "".
        name (str, optional): Name of circuit. Defaults to "two_input_one_bit_circuit".
    """
    def __init__(self, a: Wire = Wire(name="a"), b: Wire = Wire(name="b"), prefix: str = "", name: str = "two_input_one_bit_circuit"):
        super().__init__(inputs=[a, b], prefix=prefix, name=name, out_N=1, one_bit_circuit=True)
        self.c_data_type = "uint8_t"

    """ C CODE GENERATION """
    # FLAT C #
    def get_declaration_c_flat(self):
        """Generates flat C code declaration of output wires.

        Returns:
            str: Flat C code containing unique declaration of circuit wires.
        """
        # Unique declaration of all circuit's inner components outputs
        return "".join([c.out.get_declaration_c() for c in self.components if c.disable_generation is False])

    # Wires values initialization and assignment
    def get_init_c_flat(self):
        """Generates flat C code initialization and assignment of corresponding two input one bit circuit's output wires.

        Returns:
            str: Flat C code initialization of two input one bit circuit output wires.
        """
        return "".join([c.get_assign_c_flat() for c in self.components])

    # HIERARCHICAL C #
    # Subcomponent generation
    def get_function_block_c(self):
        """Generates hierarchical C code representation of corresponding two input one bit circuit used as function block in hierarchical circuit description.

        Returns:
            str: Hierarchical C code of two input one bit circuit's function block description.
        """
        adder_block = self.__class__()
        return f"{adder_block.get_circuit_c()}\n\n"

    def get_out_invocation_c(self):
        """Generates hierarchical C code invocation of corresponding two input one bit circuit's generated function block.

        Assigns output values from invocation of the corresponding function block into inner wires present inside the upper
        component from which function block has been invoked.

        Returns:
            str: Hierarchical C code subcomponent's C function invocation and output assignment.
        """
        # Used to retrieve proper component's output wire offset position within the output bus
        output_bus_wire_names = []
        [output_bus_wire_names.append(w.prefix) for w in self.out.bus]
        circuit_block = self.__class__()
        return "".join([f"  {c.out.prefix} = ({circuit_block.prefix}({self.a.get_wire_value_c_hier()}, {self.b.get_wire_value_c_hier()}) >> {output_bus_wire_names.index(c.out.prefix)}) & 0x01;\n" for c in self.components if c.disable_generation is False and c.out.prefix in output_bus_wire_names])

    # Self circuit hierarchical generation
    def get_declaration_c_hier(self):
        """Generates hierarchical C code declaration of output wires.

        Returns:
            str: Hierarchical C code containing unique declaration of circuit wires.
        """
        # Unique declaration of all circuit's inner components outputs
        return "".join([c.out.get_declaration_c() for c in self.components if c.disable_generation is False and any(o.prefix == c.out.prefix for o in self.out.bus)])

    def get_function_out_c_hier(self):
        """Generates hierarchical C code assignment of corresponding two input one bit circuit's output wires.

        Returns:
            str: Hierarchical C code containing output bus wires assignment.
        """
        return self.out.return_bus_wires_values_c_flat()

    """ VERILOG CODE GENERATION """
    # FLAT VERILOG #
    def get_prototype_v(self):
        """Generates Verilog for flat module header to describe corresponding two input one bit circuit's interface in Verilog.

        It is adapted for generation of general description of two input one bit circuits as well as their modified versions when some inputs are desired as constant values.
        In such cases the inner module logic is also modified accordingly. It is used for self two input one bit circuit flat generation.

        Returns:
            str: Flat module's name and parameters in Verilog.
        """
        unique_out_wires = []
        [unique_out_wires.append(o.name+"_outid"+str(self.out.bus.index(o))) if o.is_const() or o.name in [self.a.name, self.b.name] else unique_out_wires.append(o.name) for o in self.out.bus]
        return f"module {self.prefix}(input {self.a.name}, input {self.b.name}" + \
               "".join([f", output {o}" for o in unique_out_wires]) + ");\n"

    def get_declarations_v_flat(self):
        """Generates flat Verilog code declaration of self one bit circuit wires when described one bit circuit is a top module.

        Returns:
            str: Flat Verilog code containing unique declaration of one bit circuit wires.
        """
        # Unique declaration of all circuit's inner component wires
        return "".join([c.out.get_declaration_v_flat() for c in self.components if c.disable_generation is False and not any(o.prefix == c.out.prefix for o in self.out.bus)])

    def get_declaration_v_flat(self):
        """Generates flat Verilog code declaration of inner component's output wires.

        Returns:
            str: Flat Verilog code containing unique declaration of one bit circuit wires.
        """
        # Unique declaration of all circuit's inner components outputs
        return "".join([c.out.get_declaration_v_flat() for c in self.components if c.disable_generation is False])

    # Wires values initialization and assignment
    def get_init_v_flat(self):
        """Generates flat Verilog code initialization and assignment of corresponding two input one bit circuit's input/output wires.

        Returns:
            str: Flat Verilog code initialization of two input one bit circuit wires.
        """
        return "".join([c.get_assign_v_flat() for c in self.components])

    def get_self_init_v_flat(self):
        """Generates Verilog for self flat module initialization and assignment of corresponding two input one bit circuit's input/output wires.

        It is adapted for generation of general description of two input one bit circuits as well as their modified versions when some inputs are desired as constant values.
        In such cases the inner module prototype is also modified accordingly. It is used for self two input one bit circuit flat generation.

        Returns:
            str: Verilog flat module's inner circuit wires initialization and assignment.
        """
        unique_out_wires = []
        [unique_out_wires.append(o.name+"_outid"+str(self.out.bus.index(o))) if o.is_const() or o.name in [self.a.name, self.b.name] else unique_out_wires.append(o.name) for o in self.out.bus]
        return "".join([c.get_assign_v_flat() if c.disable_generation is False else
                        f"  assign {unique_out_wires.pop(unique_out_wires.index(c.out.name+'_outid'+str(c.outid)))} = {c.out.v_const};\n" if f"{c.out.name+'_outid'+str(c.outid)}" in unique_out_wires and c.out.is_const() else
                        f"  assign {unique_out_wires.pop(unique_out_wires.index(c.out.name+'_outid'+str(c.outid)))} = {c.out.name};\n" if f"{c.out.name+'_outid'+str(c.outid)}" in unique_out_wires else
                        f"" for c in self.components])

    # Generating flat Verilog code representation of circuit
    def get_v_code_flat(self, file_object):
        """Generates flat Verilog code representation of corresponding one bit circuit.

        Args:
            file_object (TextIOWrapper): Destination file object where circuit's representation will be written to.
        """
        file_object.write(self.get_prototype_v())
        file_object.write(self.get_declarations_v_flat())
        file_object.write(self.get_self_init_v_flat())
        file_object.write(f"endmodule")

    # HIERARCHICAL VERILOG #
    # Subcomponent generation
    def get_function_block_v(self):
        """Generates hierarchical Verilog code representation of corresponding two input one bit circuit used as function block in hierarchical circuit description.

        Returns:
            str: Hierarchical Verilog code of two input one bit circuit's function block description.
        """
        adder_block = self.__class__()
        return f"{adder_block.get_circuit_v()}\n\n"

    def get_out_invocation_v(self):
        """Generates hierarchical Verilog code invocation of corresponding two input one bit circuit's generated function block.

        Assigns output values from invocation of the corresponding function block into inner wires present inside the upper
        component from which function block has been invoked.

        Returns:
            str: Hierarchical Verilog code subcomponent's module invocation and output assignment.
        """
        circuit_block = self.__class__()
        return f"  {circuit_block.prefix} {circuit_block.prefix}_{self.out.prefix}(.{circuit_block.a.prefix}({self.a.get_wire_value_v_hier()}), .{circuit_block.b.prefix}({self.b.get_wire_value_v_hier()}){self.out.get_unique_assign_out_wires_v(circuit_block)});\n"

    # Self circuit hierarchical generation
    def get_prototype_v_hier(self):
        """Generates Verilog for hierarchical module header to describe corresponding two input one bit circuit's interface in Verilog.

        It is adapted for generation of general description of two input one bit circuits as well as their modified versions when some inputs are desired as constant values.
        In such cases the inner module logic is also modified accordingly. It is used for self two input one bit circuit hierarchical generation.

        Returns:
            str: Hierarchical module's name and parameters in Verilog.
        """
        unique_out_wires = []
        [unique_out_wires.append(o.name+"_outid"+str(self.out.bus.index(o))) if o.is_const() or o.name in [self.a.name, self.b.name] else unique_out_wires.append(o.name) for o in self.out.bus]
        return f"module {self.prefix}(input [0:0] {self.a.name}, input [0:0] {self.b.name}" + \
               "".join([f", output [0:0] {o}" for o in unique_out_wires]) + ");\n"

    def get_declarations_v_hier(self):
        """Generates hierarchical Verilog code declaration of input subcomponent's circuit wires.

        Returns:
            str: Hierarchical Verilog code containing unique declaration of subcomponent's function block circuit wires.
        """
        # Unique declaration of all circuit's inner component wires
        return "".join([c.out.get_declaration_v_hier() for c in self.components if c.disable_generation is False and not any(o.prefix == c.out.prefix for o in self.out.bus)])

    def get_declaration_v_hier(self):
        """Generates hierarchical Verilog code declaration of output wires.

        Returns:
            str: Hierarchical Verilog code containing unique declaration of circuit wires.
        """
        # Unique declaration of all circuit's inner components outputs
        return "".join([c.out.get_declaration_v_hier() for c in self.components if c.disable_generation is False and any(o.prefix == c.out.prefix for o in self.out.bus)])

    def get_self_init_v_hier(self):
        """Generates Verilog for hierarchical module's inner initialization and assignment of corresponding arithmetic circuit's input/output wires.

        It is adapted for generation of general description of two input one bit circuits as well as their modified versions when some inputs are desired as constant values.
        In such cases the inner module prototype is also modified accordingly. It is used for self two input one bit circuit hierarchical generation.

        Returns:
            str: Verilog hierarchical module's inner circuit wires initialization and assignment.
        """
        unique_out_wires = []
        [unique_out_wires.append(o.name+"_outid"+str(self.out.bus.index(o))) if o.is_const() or o.name in [self.a.name, self.b.name] else unique_out_wires.append(o.name) for o in self.out.bus]
        return "".join([c.get_gate_invocation_v() if c.disable_generation is False else
                        f"  assign {unique_out_wires.pop(unique_out_wires.index(c.out.name+'_outid'+str(c.outid)))}[0] = {c.out.v_const};\n" if f"{c.out.name+'_outid'+str(c.outid)}" in unique_out_wires and c.out.is_const() else
                        f"  assign {unique_out_wires.pop(unique_out_wires.index(c.out.name+'_outid'+str(c.outid)))}[0] = {c.out.name}[0];\n" if f"{c.out.name+'_outid'+str(c.outid)}" in unique_out_wires else
                        f"" for c in self.components])

    def get_circuit_v(self):
        """Generates hierarchical Verilog code subcomponent's function block.

        Returns:
            str: Hierarchical Verilog code of subcomponent's function block.
        """
        return f"{self.get_prototype_v_hier()}" + \
               f"{self.get_declarations_v_hier()}" + \
               f"{self.get_self_init_v_hier()}" + \
               f"endmodule"

    """ BLIF CODE GENERATION """
    # FLAT BLIF #
    def get_declaration_blif(self):
        """Generates Blif code declaration of two input one bit circuit's input/output wires.

        It is adapted for generation of general description of two input one bit circuits as well as their modified versions when some inputs are desired as constant values.
        In such cases the inner modul logic is also modified accordingly. It is used for self two input one bit circuit flat/hierarchical generation.

        Returns:
            str: Blif code containing declaration of circuit's input/output wires.
        """
        unique_out_wires = []
        [unique_out_wires.append(o.name+"_outid"+str(self.out.bus.index(o))) if o.is_const() or o.name in [self.a.name, self.b.name] else unique_out_wires.append(o.name) for o in self.out.bus]
        return f".inputs {self.a.get_wire_declaration_blif()}{self.b.get_wire_declaration_blif()}\n" + \
               f".outputs" + \
               "".join([f" {o}" for o in unique_out_wires]) + "\n" + \
               f".names vdd\n1\n" + \
               f".names gnd\n0\n"

    def get_function_blif_flat(self, top_modul: bool = False):
        """Generates flat Blif code with invocation of subcomponents logic gates Boolean functions via their corresponding truth tables.

        It is adapted for generation of general description of two input one bit circuits as well as their modified versions when some inputs are desired as constant values.
        In such cases the inner modul prototype is also modified accordingly. It is used for self two input one bit circuit flat generation.

        Args:
            top_modul (bool, optional): Specifies whether the described circuit represents top modul component (self one bit circuit generation). Defaults to False.

        Returns:
            str: Flat Blif code containing invocation of inner subcomponents logic gates Boolean functions.
        """
        if top_modul:
            unique_out_wires = []
            [unique_out_wires.append(o.name+"_outid"+str(self.out.bus.index(o))) if o.is_const() or o.name in [self.a.name, self.b.name] else unique_out_wires.append(o.name) for o in self.out.bus]
            return "".join([c.get_function_blif_flat() if c.disable_generation is False else
                           c.out.get_assign_blif(prefix=f"{unique_out_wires.pop(unique_out_wires.index(c.out.name+'_outid'+str(c.outid)))}", output=True) if f"{c.out.name+'_outid'+str(c.outid)}" in unique_out_wires else
                           "" for c in self.components])
        else:
            return "".join([c.get_function_blif_flat() for c in self.components])

    # Generating flat BLIF code representation of circuit
    def get_blif_code_flat(self, file_object):
        """Generates flat Blif code representation of corresponding one bit circuit.

        Args:
            file_object (TextIOWrapper): Destination file object where circuit's representation will be written to.
        """
        file_object.write(self.get_prototype_blif())
        file_object.write(self.get_declaration_blif())
        file_object.write(self.get_function_blif_flat(top_modul=True))
        file_object.write(f".end\n")

    # HIERARCHICAL BLIF #
    # Subcomponent/self circuit generation
    def get_function_block_blif(self):
        """Generates hierarchical Blif code representation of corresponding two input one bit circuit used as function block in hierarchical circuit description.

        Returns:
            str: Hierarchical Blif code of two input one bit circuit's function block description.
        """
        adder_block = self.__class__()
        return f"{adder_block.get_circuit_blif()}"

    def get_circuit_blif(self):
        """Generates hierarchical Blif code subcomponent's function block.

        Returns:
            str: Hierarchical Blif code of subcomponent's function block.
        """
        return f"{self.get_prototype_blif()}" + \
               f"{self.get_declaration_blif()}" + \
               f"{self.get_invocations_blif_hier()}" + \
               f".end\n"

    def get_invocations_blif_hier(self):
        """Generates hierarchical Blif code with invocation of subcomponents function blocks.

        It is adapted for generation of general description of two input one bit circuits as well as their modified versions when some inputs are desired as constant values.
        In such cases the inner modul prototype is also modified accordingly. It is used for self two input one bit circuit hierarchical generation.

        Returns:
            str: Hierarchical Blif code containing invocation of inner subcomponents function blocks.
        """
        unique_out_wires = []
        [unique_out_wires.append(o.name+"_outid"+str(self.out.bus.index(o))) if o.is_const() or o.name in [self.a.name, self.b.name] else unique_out_wires.append(o.name) for o in self.out.bus]
        return "".join([c.get_invocation_blif_hier() if c.disable_generation is False else
                       c.out.get_assign_blif(prefix=f"{unique_out_wires.pop(unique_out_wires.index(c.out.name+'_outid'+str(c.outid)))}", output=True) if f"{c.out.name+'_outid'+str(c.outid)}" in unique_out_wires else
                       "" for c in self.components])

    def get_invocation_blif_hier(self):
        """Generates hierarchical Blif code invocation of corresponding two input one bit circuit's generated function block.

        Returns:
            str: Hierarchical Blif code subcomponent's model invocation.
        """
        circuit_block = self.__class__()
        return f".subckt {circuit_block.prefix} {circuit_block.a.prefix}={self.a.get_wire_value_blif()} {circuit_block.b.prefix}={self.b.get_wire_value_blif()}{self.out.get_unique_assign_out_wires_blif(function_block_out_bus=circuit_block.out)}\n"

    """ CGP CODE GENERATION """
    # FLAT CGP #
    def get_parameters_cgp(self):
        """Generates CGP chromosome parameters of corresponding two input one bit circuit.

        In total seven parameters represent: total inputs, total outputs, number of rows, number of columns (gates),
        number of each gate's inputs, number of each gate's outputs, quality constant value.

        Returns:
            str: CGP chromosome parameters of described circuit.
        """
        self.circuit_gates = self.get_circuit_gates()
        return f"{{2,{self.out.N},1,{len(self.circuit_gates)},2,1,0}}"



from ariths_gen.wire_components.wires import ConstantWireValue0
from ariths_gen.core.one_bit_circuits import FourInputOneBitCircuit
from ariths_gen.one_bit_circuits.logic_gates import AndGate, NandGate, OrGate, NorGate, XorGate, XnorGate, NotGate
from ariths_gen.wire_components import Wire, Bus


class BlackCell(FourInputOneBitCircuit):
    """Class representing four input black cell used in parallel prefix adders inside the PG (parallel prefix computation) logic.

    Black cell is based on a notation used in the book CMOS VLSI Design (please refer to the book for more details not deducible from the code itself).
    ```
        ┌──────┐
    ───►│      ├─► Generate
    ───►│      │
    ───►│      │
    ───►│      ├─► Propagate
        └──────┘
    ```

    Description of the __init__ method.

    Args:
        a (Wire, optional): First input wire, represents generate signal from the current stage. Defaults to Wire(name="g1").
        b (Wire, optional): Second input wire, represents propagate signal from the current stage. Defaults to Wire(name="p1").
        c (Wire, optional): Third input wire, represents generate signal from a preceding stage. Defaults to Wire(name="g0").
        d (Wire, optional): Fourth input wire, represents propagate signal from a preceding stage. Defaults to Wire(name="p0").
        prefix (str, optional): Prefix name of grey cell. Defaults to "".
        name (str, optional): Name of grey cell. Defaults to "gc".
    """
    def __init__(self, a: Wire = Wire(name="g1"), b: Wire = Wire(name="p1"), c: Wire = Wire(name="g0"), d: Wire = Wire(name="p0"), prefix: str = "", name: str = "bc"):
        super().__init__(a, b, c, d, prefix=prefix, name=name)
        # 2 wires for component's bus output (generate, propagate)
        self.out = Bus(self.prefix+"_out", 2)

        # Create/propagate the generate signal
        self.add_component(AndGate(a=self.b, b=self.c, prefix=self.prefix+"_and"+str(self.get_instance_num(cls=AndGate)), parent_component=self))
        self.add_component(OrGate(a=self.a, b=self.get_previous_component().out, prefix=self.prefix+"_or"+str(self.get_instance_num(cls=OrGate)), parent_component=self))
        # Connection of the generate output wire
        self.out.connect(0, self.get_previous_component().out)

        # Create/propagate the propagate signal
        self.add_component(AndGate(a=self.b, b=self.d, prefix=self.prefix+"_and"+str(self.get_instance_num(cls=AndGate)), parent_component=self))
        # Connection of the propagate output wire
        self.out.connect(1, self.get_previous_component().out)

    def get_generate_wire(self):
        """Get output wire carrying generate signal value.

        Returns:
           Wire: Return generate wire.
        """
        return self.out.get_wire(0)

    def get_propagate_wire(self):
        """Get output wire carrying propagate signal value.

        Returns:
           Wire: Return propagate wire.
        """
        return self.out.get_wire(1)




from ariths_gen.multi_bit_circuits.others.popcount import (
    UnsignedPopCount
)

from ariths_gen.multi_bit_circuits.others.bit_reduce import (
    BitReduce,
    AndReduce,
    OrReduce
)

from ariths_gen.multi_bit_circuits.others.compare import (
    UnsignedCompareLT,
    UnsignedCompareLTE,
    UnsignedCompareGT,
    UnsignedCompareGTE
)

from ariths_gen.multi_bit_circuits.others.popcount_compare import (
    PopCountCompare
)



from ariths_gen.wire_components import (
    ConstantWireValue0,
    Bus
)
from ariths_gen.core.arithmetic_circuits import (
    GeneralCircuit
)
from ariths_gen.core.logic_gate_circuits import (
    MultipleInputLogicGate
)
from ariths_gen.one_bit_circuits.one_bit_components import (
    PGLogicBlock
)
from ariths_gen.one_bit_circuits.logic_gates import (
    AndGate,
    OrGate,
    XorGate
)


class UnsignedCarryLookaheadAdder(GeneralCircuit):
    """Class representing unsigned carry-lookahead adder.

    Unsigned carry-lookahead adder represents faster adder circuit which is composed
    of more complex circuitry but provides much less propagation delay as opposed to rca.
    It is mainly composed of propagate/generate blocks and many AND/OR gates to calculate carries individually.

    Class enables to choose the size of composite inner cla blocks (default is 4). These cla blocks are then cascaded
    to form the final N bit adder.

    ```
      B3 A3      B2 A2      B1 A1      B0 A0
      │  │       │  │       │  │       │  │
    ┌─▼──▼─┐   ┌─▼──▼─┐   ┌─▼──▼─┐   ┌─▼──▼─┐
    │  PG  │   │  PG  │   │  PG  │   │  PG  │
    │ block│   │ block│   │ block│   │ block│
    │      │   │      │   │      │   │      │
    └─┬┬┬──┘   └─┬┬┬──┘   └─┬┬┬──┘   └─┬┬┬──┘
      │││G3P3S3  │││G2P2S2  │││G1P1S1  │││G0P0S0
    ┌─▼▼▼────────▼▼▼────────▼▼▼────────▼▼▼──┐
    │     Carry Lookahead logic block       │
    │                                       │
    └┬────┬───────┬──────────┬──────────┬───┘
     │    │       │          │          │
     ▼    ▼       ▼          ▼          ▼
    Cout  S3      S1         S0         S0
    ```

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        cla_block_size (int, optional): Size of each composite cla adder block size. Defaults to 4.
        prefix (str, optional): Prefix name of unsigned cla. Defaults to "".
        name (str, optional): Name of unsigned cla. Defaults to "u_cla".
    """
    def __init__(self, a: Bus, b: Bus, cla_block_size: int = 4, prefix: str = "", name: str = "u_cla", **kwargs):
        self.N = max(a.N, b.N)
        super().__init__(inputs=[a, b], prefix=prefix, name=name, out_N=self.N+1, **kwargs)

        # Bus sign extension in case buses have different lengths
        self.a.bus_extend(N=self.N, prefix=a.prefix)
        self.b.bus_extend(N=self.N, prefix=b.prefix)

        assert cla_block_size > 0, "Block size of the individual cla groups must be greater than 0."

        # To signify current number of blocks and number of bits that remain to be added into function blocks
        block_n = 0
        N_wires = self.N
        cin = ConstantWireValue0()
        while N_wires != 0:
            # Lists containing all propagate/generate wires for the current cla block
            propagate_sig = []
            generate_sig = []
            # Cin0 used as a first generate wire for obtaining next carry bits
            generate_sig.append(cin)
            block_size = cla_block_size if N_wires >= cla_block_size else N_wires

            # Gradual addition of propagate/generate logic blocks and AND/OR gates for Cout bits generation, XOR gates for Sum bits generation
            for i in range(block_size):
                pg_block = PGLogicBlock(self.a.get_wire((block_n*cla_block_size)+i), self.b.get_wire((block_n*cla_block_size)+i), prefix=self.prefix+"_pg_logic"+str(self.get_instance_num(cls=PGLogicBlock)))
                propagate_sig.append(pg_block.get_propagate_wire())
                generate_sig.append(pg_block.get_generate_wire())
                self.add_component(pg_block)

                self.add_component(XorGate(pg_block.get_sum_wire(), cin, prefix=self.prefix+"_xor"+str(self.get_instance_num(cls=XorGate)), parent_component=self))
                self.out.connect(i+(block_n*cla_block_size), self.get_previous_component().out)

                # List of AND gates outputs that are later combined in a multi-bit OR gate
                composite_or_gates_inputs = []
                for g_index in range(len(generate_sig)-1):
                    composite_wires = []
                    # Getting a list of wires used for current bit position cout composite AND gate's generation
                    # E.g. for Cout2 = G1 + G0·P1 C0·P0·P1 it gets a list containing [C0,P0,P1] then [G0,P1]
                    composite_wires.append(generate_sig[g_index])
                    for p_index in range(len(propagate_sig)-1, g_index-1, -1):
                        composite_wires.append(propagate_sig[p_index])

                    # For each pg pair values algorithmically combine two input AND gates to replace multiple input gates (resolves fan-in issue)
                    pg_wires = Bus(wires_list=composite_wires)
                    multi_bit_and_gate = MultipleInputLogicGate(a=pg_wires, two_input_gate_cls=AndGate, prefix=self.prefix+f"_and{block_n}_{i}_{g_index}", parent_component=self)
                    composite_or_gates_inputs.append(multi_bit_and_gate.out)

                # Final OR gates cascade using generated AND gates output wires
                composite_or_wires = Bus(wires_list=composite_or_gates_inputs)
                multi_bit_or_gate = MultipleInputLogicGate(a=composite_or_wires, two_input_gate_cls=OrGate, prefix=self.prefix+f"_orred{block_n}_{i}_{g_index}_", parent_component=self)
                # Carry bit generation
                obj_cout_or = OrGate(pg_block.get_generate_wire(), multi_bit_or_gate.out, prefix=self.prefix+f"_or{block_n}_{i}_{g_index}_"+str(self.get_instance_num(cls=OrGate, count_disabled_gates=False)), parent_component=self)
                self.add_component(obj_cout_or)
                # Updating cin for the the next cla block/sum XOR
                cin = obj_cout_or.out

            N_wires -= block_size
            block_n += 1

        # Connection of final Cout
        self.out.connect(self.N, cin)


class SignedCarryLookaheadAdder(UnsignedCarryLookaheadAdder, GeneralCircuit):
    """Class representing signed carry-lookahead adder.

    Signed carry-lookahead adder represents faster adder circuit which is composed
    of more complex circuitry but provides much less propagation delay as opposed to rca.
    It is mainly composed of propagate/generate blocks and many AND/OR gates to calculate carries individually.

    Class enables to choose the size of composite inner cla blocks (default is 4). These cla blocks are then cascaded
    to form the final N bit adder.
    At last XOR gates are used to ensure proper sign extension.

    ```
      B3 A3      B2 A2      B1 A1      B0 A0
      │  │       │  │       │  │       │  │
    ┌─▼──▼─┐   ┌─▼──▼─┐   ┌─▼──▼─┐   ┌─▼──▼─┐
    │  PG  │   │  PG  │   │  PG  │   │  PG  │
    │ block│   │ block│   │ block│   │ block│
    │      │   │      │   │      │   │      │
    └─┬┬┬──┘   └─┬┬┬──┘   └─┬┬┬──┘   └─┬┬┬──┘
      │││G3P3S3  │││G2P2S2  │││G1P1S1  │││G0P0S0
    ┌─▼▼▼────────▼▼▼────────▼▼▼────────▼▼▼──┐
    │         Carry Lookahead logic         │
    │          with sign extension          │
    └┬────┬───────┬──────────┬──────────┬───┘
     │    │       │          │          │
     ▼    ▼       ▼          ▼          ▼
    Cout  S3      S1         S0         S0
    ```

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        cla_block_size (int, optional): Size of each composite cla adder block size. Defaults to 4.
        prefix (str, optional): Prefix name of signed cla. Defaults to "".
        name (str, optional): Name of signed cla. Defaults to "s_cla".
    """
    def __init__(self, a: Bus, b: Bus, cla_block_size: int = 4, prefix: str = "", name: str = "s_cla", **kwargs):
        N = max(a.N, b.N)
        a.bus_extend(N=N, prefix=a.prefix, desired_extension_wire=a.get_wire(a.N-1))
        b.bus_extend(N=N, prefix=b.prefix, desired_extension_wire=b.get_wire(b.N-1))

        super().__init__(a=a, b=b, cla_block_size=cla_block_size, prefix=prefix, name=name, signed=True, **kwargs)

        # Additional XOR gates to ensure correct sign extension in case of sign addition
        sign_xor_1 = XorGate(self.a.get_wire(self.N-1), self.b.get_wire(self.N-1), prefix=self.prefix+"_xor"+str(self.get_instance_num(cls=XorGate)), parent_component=self)
        self.add_component(sign_xor_1)
        sign_xor_2 = XorGate(sign_xor_1.out, self.get_previous_component(2).out, prefix=self.prefix+"_xor"+str(self.get_instance_num(cls=XorGate)), parent_component=self)
        self.add_component(sign_xor_2)
        self.out.connect(self.N, sign_xor_2.out)



import os
import sys
# Add the parent directory to the system path
DIR_PATH = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, os.path.join(DIR_PATH, '..'))

import numpy as np
import math
from io import StringIO

from ariths_gen.wire_components import (
    Wire,
    ConstantWireValue0,
    ConstantWireValue1,
    Bus
)

from ariths_gen.core.arithmetic_circuits import GeneralCircuit

from ariths_gen.multi_bit_circuits.others import (
    UnsignedCompareLT, UnsignedCompareLTE, UnsignedCompareGT, UnsignedCompareGTE
)

from ariths_gen.core.cgp_circuit import UnsignedCGPCircuit, SignedCGPCircuit


import numpy as np
import math
from io import StringIO


def test_compare_lt_same():
    """ Test unsigned comparator """
    N = 8

    a = Bus(N=N, prefix="a")
    b = Bus(N=N, prefix="b")
    av = np.arange(2**N).reshape(1, -1)
    bv = np.arange(2**N).reshape(-1, 1)


    cmp = UnsignedCompareLT(a=a, b=b)
    cmp.get_v_code_hier(open("tmp.verilog", "w"))

    v = cmp(av, bv)
    print("ret = ", v)

    expected = np.array(av < bv).astype(int)

    print("expected = ", expected)
        
    #expected = np.sum(r, axis=1)

    np.testing.assert_array_equal(v, expected)


    

def test_compare_lt_small():
    """ Test unsigned comparator """
    N = 8

    a = Bus(N=N, prefix="a")
    b = Bus(N=N//2, prefix="b")
    av = np.arange(2**N).reshape(1, -1)
    bv = np.arange(2**(N //2)).reshape(-1, 1)


    cmp = UnsignedCompareLT(a=a, b=b)
    #o = StringIO()
    cmp.get_python_code_flat(open("tmp.py", "w"))
    cmp.get_c_code_flat(open("tmp.c", "w"))
    cmp.get_cgp_code_flat(open("tmp.cgp", "w"))
    #cmp.get_v_code_hier(open("tmp.verilog", "w"))
    #print(o.getvalue())

#    av = 0
 #   bv = 5


    v = cmp(av, bv)
    print("ret = ", v)

    expected = np.array(av < bv).astype(int)

    print("expected = ", expected)
        
    #expected = np.sum(r, axis=1)

    np.testing.assert_array_equal(v, expected)



    
def test_compare_lte_same():
    """ Test unsigned comparator """
    N = 8

    a = Bus(N=N, prefix="a")
    b = Bus(N=N, prefix="b")
    av = np.arange(2**N).reshape(1, -1)
    bv = np.arange(2**N).reshape(-1, 1)


    cmp = UnsignedCompareLTE(a=a, b=b)
    cmp.get_v_code_hier(open("tmp.verilog", "w"))

    v = cmp(av, bv)
    print("ret = ", v)

    expected = np.array(av <= bv).astype(int)

    print("expected = ", expected)
        
    #expected = np.sum(r, axis=1)

    np.testing.assert_array_equal(v, expected)


    

def test_compare_lte_small():
    """ Test unsigned comparator """
    N = 8

    a = Bus(N=N, prefix="a")
    b = Bus(N=N//2, prefix="b")
    av = np.arange(2**N).reshape(1, -1)
    bv = np.arange(2**(N //2)).reshape(-1, 1)


    cmp = UnsignedCompareLTE(a=a, b=b)
    #o = StringIO()
    cmp.get_python_code_flat(open("tmp.py", "w"))
    cmp.get_c_code_flat(open("tmp.c", "w"))
    cmp.get_cgp_code_flat(open("tmp.cgp", "w"))
    #cmp.get_v_code_hier(open("tmp.verilog", "w"))
    #print(o.getvalue())

#    av = 0
 #   bv = 5


    v = cmp(av, bv)
    print("ret = ", v)

    expected = np.array(av <= bv).astype(int)

    print("expected = ", expected)
        
    #expected = np.sum(r, axis=1)

    np.testing.assert_array_equal(v, expected)


def test_compare_lte_small2():
    """ Test unsigned comparator """
    N = 8

    a = Bus(N=N//2, prefix="a")
    b = Bus(N=N, prefix="b")
    av = np.arange(2**(N // 2)).reshape(1, -1)
    bv = np.arange(2**(N)).reshape(-1, 1)


    cmp = UnsignedCompareLTE(a=a, b=b)
    #o = StringIO()
    cmp.get_python_code_flat(open("tmp.py", "w"))
    cmp.get_c_code_flat(open("tmp.c", "w"))
    cmp.get_cgp_code_flat(open("tmp.cgp", "w"))
    #cmp.get_v_code_hier(open("tmp.verilog", "w"))
    #print(o.getvalue())

#    av = 0
 #   bv = 5


    v = cmp(av, bv)
    print("ret = ", v)

    expected = np.array(av <= bv).astype(int)

    print("expected = ", expected)
        
    #expected = np.sum(r, axis=1)

    np.testing.assert_array_equal(v, expected)


def test_compare_gt_same():
    """ Test unsigned comparator """
    N = 8

    a = Bus(N=N, prefix="a")
    b = Bus(N=N, prefix="b")
    av = np.arange(2**N).reshape(1, -1)
    bv = np.arange(2**N).reshape(-1, 1)


    cmp = UnsignedCompareGT(a=a, b=b)
    cmp.get_v_code_hier(open("tmp.verilog", "w"))

    v = cmp(av, bv)
    print("ret = ", v)

    expected = np.array(av > bv).astype(int)

    print("expected = ", expected)
        
    #expected = np.sum(r, axis=1)

    np.testing.assert_array_equal(v, expected)


    

def test_compare_gt_small():
    """ Test unsigned comparator """
    N = 8

    a = Bus(N=N, prefix="a")
    b = Bus(N=N//2, prefix="b")
    av = np.arange(2**N).reshape(1, -1)
    bv = np.arange(2**(N //2)).reshape(-1, 1)


    cmp = UnsignedCompareGT(a=a, b=b)
    #o = StringIO()
    cmp.get_python_code_flat(open("tmp.py", "w"))
    cmp.get_c_code_flat(open("tmp.c", "w"))
    cmp.get_cgp_code_flat(open("tmp.cgp", "w"))
    #cmp.get_v_code_hier(open("tmp.verilog", "w"))
    #print(o.getvalue())

#    av = 0
 #   bv = 5


    v = cmp(av, bv)
    print("ret = ", v)

    expected = np.array(av > bv).astype(int)

    print("expected = ", expected)
        
    #expected = np.sum(r, axis=1)

    np.testing.assert_array_equal(v, expected)


    
def test_compare_gte_same():
    """ Test unsigned comparator """
    N = 8

    a = Bus(N=N, prefix="a")
    b = Bus(N=N, prefix="b")
    av = np.arange(2**N).reshape(1, -1)
    bv = np.arange(2**N).reshape(-1, 1)


    cmp = UnsignedCompareGTE(a=a, b=b)
    cmp.get_v_code_hier(open("tmp.verilog", "w"))

    v = cmp(av, bv)
    print("ret = ", v)

    expected = np.array(av >= bv).astype(int)

    print("expected = ", expected)
        
    #expected = np.sum(r, axis=1)

    np.testing.assert_array_equal(v, expected)


    

def test_compare_gte_small():
    """ Test unsigned comparator """
    N = 8

    a = Bus(N=N, prefix="a")
    b = Bus(N=N//2, prefix="b")
    av = np.arange(2**N).reshape(1, -1)
    bv = np.arange(2**(N //2)).reshape(-1, 1)


    cmp = UnsignedCompareGTE(a=a, b=b)
    #o = StringIO()
    cmp.get_python_code_flat(open("tmp.py", "w"))
    cmp.get_c_code_flat(open("tmp.c", "w"))
    cmp.get_cgp_code_flat(open("tmp.cgp", "w"))
    #cmp.get_v_code_hier(open("tmp.verilog", "w"))
    #print(o.getvalue())

#    av = 0
 #   bv = 5


    v = cmp(av, bv)
    print("ret = ", v)

    expected = np.array(av >= bv).astype(int)

    print("expected = ", expected)
        
    #expected = np.sum(r, axis=1)

    np.testing.assert_array_equal(v, expected)



def test_compare_gte_cgp_same():
    """ Test unsigned comparator """
    N = 8

    a = Bus(N=N, prefix="a")
    b = Bus(N=N, prefix="b")
    av = np.arange(2**N).reshape(1, -1)
    bv = np.arange(2**N).reshape(-1, 1)


    cmp = UnsignedCompareGTE(a=a, b=b)
    o = StringIO()
    cmp.get_v_code_hier(open("tmp.verilog", "w"))
    cmp.get_cgp_code_flat(o)
    
    cgp = UnsignedCGPCircuit(o.getvalue(), [N, N])

    v = cgp(av, bv)
    print("ret = ", v)

    expected = np.array(av >= bv).astype(int)

    print("expected = ", expected)
        
    #expected = np.sum(r, axis=1)

    np.testing.assert_array_equal(v, expected)


    

def test_compare_gte_cgp_small():
    """ Test unsigned comparator """
    N = 8

    a = Bus(N=N, prefix="a")
    b = Bus(N=N//2, prefix="b")
    av = np.arange(2**N).reshape(1, -1)
    bv = np.arange(2**(N //2)).reshape(-1, 1)


    cmp = UnsignedCompareGTE(a=a, b=b)
    o = StringIO()
    cmp.get_v_code_flat(open("tmp.verilog", "w"))
    cmp.get_cgp_code_flat(o)
    cmp.get_cgp_code_flat(open("tmp.cgp", "w"))
    
    cgp = UnsignedCGPCircuit(o.getvalue(), [N, N // 2])

    v = cgp(av, bv)
    #cmp.get_v_code_hier(open("tmp.verilog", "w"))
    #print(o.getvalue())

#    av = 0
 #   bv = 5


    print("ret = ", v)

    expected = np.array(av >= bv).astype(int)

    print("expected = ", expected)
        
    #expected = np.sum(r, axis=1)

    np.testing.assert_array_equal(v, expected)
